/* INC-REG.c (version 2). This is an implementation of the INC-REG algorithm from A. Arman, P. Gao and N. Wormald,
 * Fast uniform generation of random graphs with given degree sequences, 2019 IEEE 60th Annual Symposium on
 * Foundations of Computer Science (FOCS) 1,  1371-1379 (2019).
 * The algorithm generates a uniformly random regular graph with given degree and number of vertices.
 * Original code written by Alon Naor. Enquiries: nicholas.wormald@monash.edu
 */

/*
 Version information:
 
 2020/08/09 v2: improved i/o
 
 */

/******************************************************************************************************************/
/********************************************   LICENCE INFORMATION   *********************************************/
/******************************************************************************************************************/

/*
 
 Copyright 2020 Monash University
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 
 http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdbool.h>
#include <string.h>
#include <sys/time.h>
#include <inttypes.h>
#include <limits.h>



/******************************************************************************************************************/
/**********************************************   USER DEFINITIONS   **********************************************/
/******************************************************************************************************************/

/*********************************   RNG  *********************************/
/* Define EXACTLY one of the following to select the (pseudo) Random Number Generator (RNG).
 * 1. Define SIMPLE_RNG to use sequential calls to C standard function rand.
 * 2. Define LONG_CYCLE_RNG to use the algorithm from Press, Teukolsky, Vetterling and Flannery, Numerical Recipes. The art of scientific computing.
*/
#define SIMPLE_RNG
//#define LONG_CYCLE_RNG

/* If USE_FIXED_SEED is set to 1, initial seed is FIXED_SEED, otherwise it is random.
 * Note that SIMPLE_RNG requires a non-negative seed, whereas
 * LONG_CYCLE_RNG is only initialized if the seed is non-positive.
 */
#define USE_FIXED_SEED 0
#define FIXED_SEED 1234
/**************************************************************************/



/***************************   GRAPHS OUTPUT   ****************************/
// Set DO_NOT_OUTPUT to 1 if you are not interested in the actual output of the generated graphs.
#define DO_NOT_OUTPUT 0

/* Determines the graph output type.
 * If OUTPUT_GRAPH_BY_VERTEX is set to 0, the output is a list of all edges in the graph.
 * Otherwise, for each vertex its list of neighbors of larger index is printed.			*/
#define OUTPUT_GRAPH_BY_VERTEX 1

/* Determines whether the graph output is sorted or not.
 * Applies for both edge list and neighbors lists output. */
#define OUTPUT_GRAPH_SORTED 1

// For neighbors lists output, determines whether vertices with empty lists are printed.
#define PRINT_EMPTY_LISTS 0

// For edge list output, determines how many edges are printed per line.
#define EDGES_PER_LINE 10

/* For edge list output, determines the column width of the output.
 * When set to 0, width is calibrated automatically according to n. */
#define COLUMN_WIDTH 0
/**************************************************************************/



/*****************************   PERFORMANCE   ****************************/
/* The hash size of any vertex is degree * HASH_MULTIPLIER (rounded up).
 * The smaller this value is, less memory is allocated and larger graphs can be handled,
 * but more collisions will occur in the hash, resulting in slower insert/find operations.
 * The value can be any positive real number.
 */
#define HASH_MULTIPLIER 1

/* Determines the data type used for M, the sum of degrees (which affects the
 * amount of allocated memory). Set to 1 if M >= 2^31, otherwise set to 0.
 */
#define M_64 0

/* Adds safety checks such as data structures verification and brute
 * count validation for the values computed for backwards rejection.
 * Significantly slows down runtime. Should normally be set to 0.
 */
#define SAFETY_CHECKS 0

/* When DEBUG_PRINTS is set to 1, debug data is printed to log_file.
 * Should normally be set to 0, unless working with a small input and
 * wish to examine carefully the execution of the code.
 */
#define DEBUG_PRINTS 0
/**************************************************************************/



/*********************************   MISC   *******************************/
/* The algorithm reverts to INC-GEN if gamma and epsilon cannot be found for INC-REG.
 * If set to 0, user has to approve reversion or abort in these cases; otherwise,
 * reverts automatically. The same goes for forbidding loops, triple edges or double
 * edges, which occur if the probability for backward rejection is too high for them. */
#define REVERT_AUTOMATICALLY 1

/* Sets the limit for a potential exit point after many consecutive
 * failed attempts. Set to 0 if not interested in this feature.	 */
#define CONSECUTIVE_FAILED_ATTEMPTS_LIMIT 1000

// Determines if statistics summary is printed to the log_file during the run.
#define PRINT_SUMMARY 0

/* Determines if the counter of successfully generated graphs is printed to
 * the error file during the run.
 */
#define SHOW_PROGRESS 0
/**************************************************************************/

/******************************************************************************************************************/
/****************************************   OTHER DEFINITIONS (NON-USER)  *****************************************/
/******************************************************************************************************************/

typedef uint64_t u64;
#define S16_MAX  0x7fff
#define S32_MAX  0x7fffffff

/**********************************   RNG   *******************************/
/* randInt uses repeated calls to a base RNG function and concatenates random bits.
 * Here the base call and number of bits it generates are defined according to the selected RNG.
 */
#ifdef SIMPLE_RNG
#if (RAND_MAX == S16_MAX)
// Here rand() returns 15 random bits.
#define RAND_BASE_CALL   rand()
#define NUM_OF_RAND_BITS 15
#endif
#if (RAND_MAX == S32_MAX)
// Here rand() returns 31 random bits, from which the LSB is removed.
#define RAND_BASE_CALL   (rand() / 2)
#define NUM_OF_RAND_BITS 30
#endif
// Verifies the maximum value returned by rand() is supported.
#ifndef RAND_BASE_CALL
#error Cannot use SIMPLE_RNG. Use LONG_CYCLE_RNG or add support for the value of RAND_MAX.
#endif
#endif

#ifdef LONG_CYCLE_RNG
// LRAN(1 << 30) returns 30 random bits (can't have 31 since LRAN never returns INT_MAX).
#define RAND_BASE_CALL   LRAN(1 << 30)
#define NUM_OF_RAND_BITS 30
#endif

// The maximal range of integers randInt can produce is 2^60 (i.e. 60 random bits).
#define RAND_INT_MAX (1ULL << 60)

// Preprocessor verification: enforces the definition of exactly one RNG.
#if((defined(LONG_CYCLE_RNG) && defined(SIMPLE_RNG)) || \
	(!defined(LONG_CYCLE_RNG) && !defined(SIMPLE_RNG)))
#error EXACTLY one of LONG_CYCLE_RNG and SIMPLE_RNG must be defined.
#endif
/**************************************************************************/



/*****************************   PERFORMANCE   ****************************/
// Defines the type of M and the matching print specifier according to user's selection.
#if M_64
typedef u64		 type_M;
#define PRINT_M	 PRIu64
#else
typedef uint32_t type_M;
#define PRINT_M  PRIu32
#endif /* M_64 */

#if SAFETY_CHECKS
// Makes the data structures verification calls invoke the actual functions.
#define checkPointsArray()		_checkPointsArray()
#define verifyDataStructures()  _verifyDataStructures()
#else
// Removes from the code all data structures verification calls.
#define checkPointsArray()
#define verifyDataStructures()
#endif /* SAFETY_CHECKS */
/**************************************************************************/



/***************************   OUTPUT / LOGGING   *************************/
// Determines the correct outputGraph function to be invoked.
#if DO_NOT_OUTPUT
#define outputGraph()
#elif OUTPUT_GRAPH_SORTED
#define outputGraph outputGraphSorted
#elif OUTPUT_GRAPH_BY_VERTEX
#define outputGraph outputGraphUnsortedNeighbors
#else
#define outputGraph outputGraphUnsortedEdges
#endif

// A default log file name if one is not provided.
#define DEFAULT_LOG_FILE "INC_REG_log.txt"

// The different types of output in the code.
#define LOG_ERR(param...)	fprintf(stderr, param)
#define LOG_INFO(param...)	fprintf(log_file, param)
#define OUTPUT(param...)	fprintf(output_file, param)

#if REVERT_AUTOMATICALLY
#define LOG_REVERT	LOG_INFO
#else
#define LOG_REVERT	LOG_ERR
#endif

#if DEBUG_PRINTS
// Enables debug logging.
#define LOG_DEBUG			 LOG_INFO
// Makes the print calls invoke the actual printing functions.
#define printPairing		 _printPairing
#define printNeighborsByHash _printNeighborsByHash
#define printAllNeighbors 	 _printAllNeighbors
#else
// Removes from the code all printing calls.
#define LOG_DEBUG(param...)
#define printPairing()
#define printNeighborsByHash(u)
#define printAllNeighbors()
#endif /* DEBUG_PRINTS */
/**************************************************************************/



/******************************************************************************************************************/
/********************************************   STRUCTURES AND ENUMS  *********************************************/
/******************************************************************************************************************/

/* <adjNode> (adjacency node)
 * Adjacency nodes are used to store the neighbors of each vertex, loops excluded.
 * For each vertex, all of its nodes are stored consecutively, and they are additionally mapped into a hash table,
 * where each hash cell contains an unordered linked list of nodes.
 * The multiplicity of the edge corresponding to a given node is indicated by the "edge" field.
 * Multiple edges are initially represented by one adjacency node (for each of the edge's endpoints), but when NoDoubles
 * is invoked a second adjacency node is created for each double edge, and the two nodes representing the same neighbor
 * are kept consecutively in the appropriate linked list (this is done for easier handling of Class B D-switches).
 *
 * Fields:
 * edge - the value in this field changes according to the multiplicity of the edge this node corresponds to:
 * 		  for a single edge, points to its location (the index) in the points array;
 * 		  for a double edge, holds some value k, where M <= k < 2 * M;
 * 		  for a triple edge, holds the value 2 * M.
 * 		  Additionally, after the random pairing of all points is generated and the points array is scanned to
 * 		  detect all loops and multiple edges, this field is used to locate the "multiEdge" data structure (see below)
 * 		  holding the information regarding the multiple edge corresponding to the node.
 * id 	- the index of the neighbor.
 * next - points to the next node in the linked list (NULL if this node is last in the list).
 */
typedef struct _adjNode{
	type_M 			 edge;
	int 			 id;
	struct _adjNode* next;
}adjNode;



/* <vertex>
 * The vertex structure holds all information of a given vertex.
 * The "hash" and "neighbors_pool" fields are fixed throughout the run,
 * while the other fields change according to the current status of the graph.
 *
 * Fields:
 * has_loop				- a boolean indicating whether there is a loop on this vertex or not.
 * single_edges_num		- the number of single edges incident with the vertex.
 * hash					- pointer to the beginning of the hash table allocated for this vertex.
 * neighbors_pool		- pointer to the beginning of the neighbors pool allocated for this vertex.
 * next_available_node	- pointer to the next available adjacency node to be allocated for this vertex.
 * 					  	  Initially points to the beginning of the section kept for this vertex in the
 * 					  	  adjacency nodes pool, then advancing with the allocation of each node.
 */
typedef struct{
	bool 	 has_loop;
	int 	 single_edges_num;
	adjNode** hash;
	adjNode* neighbors_pool;
	adjNode* next_available_node;
}vertex;



/* <multiEdge>
 * After the random pairing of all points is generated, the points array is scanned to detect all loops and multiple edges.
 * All relevant information for each detected multiple edge is stored in a multiEdge structure. When the scan of the
 * points array is complete (assuming this is a valid pairing), this information is used to rearrange the points array
 * such that the order of pairs is: loops, single edges, triple edges, double edges.
 *
 * Fields:
 * v1, v2		 - the two vertices comprising the edge.
 * first, second - the first two detected locations of the pair v1, v2.
 * third		 - the third detected location of the pair v1, v2 if this is a triple edge, and -1 otherwise.
 */
typedef struct{
	int	   v1;
	int	   v2;
	type_M first;
	type_M second;
	type_M third;
}multiEdge;



/* <jointNeighbors>
 * When computing B_d(G) and B_t(G) "bad" configurations are counted (with respect to a given 2-path or 3-star S).
 * When dealing with configurations which contribute to the count due to the presence of forbidden edges, it is
 * crucial to avoid double counting in case multiple such edges exist. This is done with the jointNeighbors structure.
 * Here v3 and v5 are specific vertices of S-degree 1, each may be the intersection of S and a forbidden edge.
 *
 * Fields:
 * joint_with_v3	- for a vertex u, the number of vertices adjacent to both u and v3.
 * joint_with_v5	- for a vertex u, the number of vertices adjacent to both u and v5.
 * joint_with_v3_v5	- for a vertex u, the number of vertices adjacent to u, v3 and v5.
 */
typedef struct{
	int joint_with_v3;
	int joint_with_v5;
	int joint_with_v3_v5;
}jointNeighbors;



/* Indicates the type (location) of existing edge(s) between the double edge (v2,v5)
 * and the two single edges (v1,v4) and (v3,v6) when performing a Type I D-switch. */
typedef enum{
	v2_v3,
	v2_v1,
	v5_v6,
	v5_v4,
	no_edge,
	multiple_edges
}existing_edge_type;



// Indicates the exit reason upon program termination.
typedef enum{
	successful_termination,
	user_termination,
	file_failure,
	input_failure,
	configuration_failure,
	memory_allocation_failure,
	safety_check_failure
}exit_value;



/******************************************************************************************************************/
/**********************************************   GLOBAL VARIABLES   **********************************************/
/******************************************************************************************************************/

/* General parameters, depending only on the degree sequence (i.e. n and d). */
int n; 				// Number of vertices in the graph.
u64	degree;			// The degree of every vertex in the graph.
					// Bounded by 2^16, but kept as u64 to avoid potential overflows in some computations.

u64 M;				// Sum of degrees		 = n * d.
u64 M2;				// Sum of FALL_2(degree) = n * d * (d-1).
u64 M3;				// Sum of FALL_3(degree) = n * d * (d-1) * (d-2).

int B_L;			// The maximum allowed number of loops.
int B_D;			// The maximum allowed number of double edges.
int B_T;			// The maximum allowed number of triple edges.

double Gamma;		// Technical parameter, used to compute B_D and epsilon.
double epsilon;		// Technical parameter, used to compute the probability of performing a Type I D-switch.
double *rhoII;		// rhoII[i] is the probability of performing a type II D-switch when double_count = i.
double *bA_approx;	// bA_approx[i] is a lower bound on the number of potential Class A D-switches when double_count = i.
					// The approximation is used for b-rejection and only if M2 > 2^30. Otherwise exact value is computed.

bool revert_to_inc_gen;	// Indicates whether INC-REG has reverted to INC-GEN (happens if gamma and epsilon could not be found).

u64 hash_cells_per_vertex;	// Number of cells in each hash table = ceiling(degree * HASH_MULTIPLIER).
							// Kept as u64 to avoid potential overflows in some computations.
int hash_cell_size;	// The number of indices mapped to each cell in the hash table = ceiling(n / hash_cells_per_vertex).
					// If this value is k, then the vertices in the range [ik, ..., (i + 1)k) are mapped to the
					// i-th hash cell (including i = 0).


/* General graph parameters which change according to the current status of the processed graph. */
int loop_count;		// Current number of loops.
int	double_count;	// Current number of double edges.
int triple_count;	// Current number of triple edges.
u64 P2;				// Current number of simple ordered 2-paths v1-v2-v3 such that there is no loop on v2.
u64 S3;				// Current number of simple ordered 3-stars [v0; v1,v2,v3].


/* The data structures holding all information for each processed graph. */
vertex 	*V;				// The array of all vertex structures.
int 	*points;		// The points array, stores all pairs in this order: loops, single edges, triple edges, double edges.
adjNode **hash_array;	// The array of all hash entries. Distributed equally among the vertices.
adjNode	*adjNode_pool;	// The array of all adjacency nodes. Distributed equally among the vertices.


/* Multiple edges data which is used during the the pairing generation to sort the points array. */
multiEdge *multi_edges;				// Stores for each multiple edge the locations of all pairs corresponding to it.
type_M 	  *multi_edges_locations;	// Stores the location of all pairs corresponding to multiple edges.


/* Data structures used when computing B_d(G) and B_t(G) to avoid double counting of "bad" configurations. */
jointNeighbors *joint_neighbors_tracker; 	// Collects the info preventing the double counting in B_d(G) and B_t(G).
int	*encountered;			// Used for efficient initialization at the end of each computation of B_d(G) and B_t(G).


/* File streams. */
FILE *output_file;	// The file to which generated graphs are printed. Set to stdout if no output file is provided.
FILE *log_file;		// The file to which the log is printed. Only applicable when DEBUG_PRINTS is set to 1.


/* RNG seed. */
int seed;			// Only applicable if LONG_CYCLE_RNG is defined. Given as a parameter in every call to LRAN.


/* Statistics variables. */
int double_loop, quadruple_edge;											// Counts pairing generations failures.
int many_loops, many_double_edges, many_triple_edges, many_multiple_edges;	// Counts pairing generations failures.

int total_L_switch_calls, total_D_switch_calls, total_T_switch_calls;		// Counts all calls to L/D/T-switches.
int L_f_rej, L_b_rej, D_f_rej, D_b_rej, T_f_rej, T_b_rej;					// Counts all f/b-rejections according to L/D/T.

int total_noloops,   total_loop_num;			// Counts calls to NoLoops and the accumulated sum of loops.
int total_nodoubles, total_double_edges_num;	// Counts calls to NoDoubles and the accumulated sum of double edges.
int total_notriples, total_triple_edges_num;	// Counts calls to NoTriples and the accumulated sum of triple edges.

int type_I_f_rej, type_II_f_rej, Class_A_b_rej, Class_B_b_rej;		// Counts f/b-rejections of D-switches according to type and class.
int total_Type_I, total_Type_II, type_rej, total_Class_A, total_Type_I_Class_B;	// Counts type and class distribution of all D-switches.



/******************************************************************************************************************/
/***************************************************   MACROS   ***************************************************/
/******************************************************************************************************************/

// Some basic utility functions.
#define IS_EVEN(num) ((num) % 2 == 0)
#define FALL_2(num)  (((u64) num) * (num - 1))
#define FALL_3(num)  (((u64) num) * (num - 1) * (num - 2))

// Checks if a given string is a number. Used for input parsing.
#define NOT_A_NUMBER(str) (strspn(str, "0123456789") < strlen(str))

// Compares two pairs of vertices, used when data structures are verified.
#define ARE_PAIRS_EQUAL(a1, b1, a2, b2)	(((a1 == a2) && (b1 == b2)) || ((a1 == b2) && (b1 == a2)))

// Swaps a and b, used for relabeling vertices in a Type-I-Class-B D-switch.
#define SIMPLE_SWAP(a,b){	\
	a ^= b;					\
	b ^= a;					\
	a ^= b;					\
}

/* Swaps two entries in the points array. Only invoked while generating
 * a new pairing and before P2 is computed so P2 is used as tmp.
 * SIMPLE_SWAP cannot be used since a might equal b.				 */
#define POINTS_SWAP(a,b){	\
	P2		  = points[a];	\
	points[a] = points[b];	\
	points[b] = P2;			\
}

// Allocates memory and exits upon failure.
#define ALLOCATE_MEMORY(name, type, num){												\
	name = NULL;																		\
	if(SIZE_MAX / sizeof(type) >= num)													\
		name = (type*) malloc (sizeof(type) * (num));									\
	if(!name){																			\
		LOG_ERR("Could not allocate memory for %s (%" PRIu64 " bytes). Exiting.\n",		\
				#name, ((u64) sizeof(type)) * (num));									\
		finalize(memory_allocation_failure);											\
	}																					\
	LOG_DEBUG("Allocated %" PRIu64 " bytes for %s.\n", 									\
			  ((u64) sizeof(type)) * (num), #name);										\
}

// Allocates memory and sets the entire block to zero.
#define ALLOCATE_INITIALIZED_MEMORY(name, type, num){	\
	ALLOCATE_MEMORY(name, type, num);					\
	memset(name, 0, sizeof(type) * (num));				\
}

// Measures and prints the runtime of a given operation (could be a sequence of operations).
#define MEASURE(str,operation){								\
	struct timeval  tv1, tv2;								\
	gettimeofday(&tv1, NULL);								\
	operation;												\
	gettimeofday(&tv2, NULL);								\
	LOG_INFO("Total time %s = %.3f seconds.\n", str,		\
			 (tv2.tv_usec - tv1.tv_usec) / 1000000.0 +		\
			 tv2.tv_sec - tv1.tv_sec);						\
}

// Retrieves the cell in the hash of 'v' to which 'neighbor' is mapped.
#define GET_CELL(v,neighbor) (V[v].hash + ((neighbor)/hash_cell_size))

// Retrieves the adjNode of 'neighbor' from the hash of 'v' if exists, and NULL otherwise.
#define GET_NODE(v,neighbor,node){					\
	node = *(GET_CELL(v,neighbor));					\
	while(node != NULL && node->id != neighbor)		\
		node = node->next; 							\
}

// Inserts the node of a new neighbor of a vertex u to the top of its hash cell.
#define HEAD_INSERT(u, node){					\
	adjNode** head = GET_CELL(u,node->id);		\
	node->next = *head;							\
	*head = node;								\
}

// Extracts the adjacency node of 'neighbor' from the hash of 'v'.
#define EXTRACT_NODE(v, neighbor, node){		\
	/* Find the previous node in the list. */	\
	adjNode **prev = GET_CELL(v, neighbor);		\
	while((*prev)->id != neighbor){				\
		prev = &((*prev)->next);				\
	}											\
	/* Extract the node. */						\
	node  = *prev;								\
	*prev = node->next;							\
}

// Queries regarding the existence or the multiplicity of an edge are wrapped in one of the following.
#define IS_EDGE(u,v)		(countEdgeMultiplicity(u,v)  > 0)
#define IS_NO_EDGE(u,v)		(countEdgeMultiplicity(u,v) == 0)
#define IS_SINGLE_EDGE(u,v) (countEdgeMultiplicity(u,v) == 1)
#define IS_DOUBLE_EDGE(u,v) (countEdgeMultiplicity(u,v) == 2)
//#define IS_TRIPLE_EDGE(u,v) (countEdgeMultiplicity(u,v) == 3)

/* Checks the multiplicity of the edge corresponding to the given node.
 * Mainly used for neighbors iterations in which only single edges should be processed. */
#define IS_NODE_OF_SINGLE_EDGE(node) (node->edge < M)
#define IS_NODE_OF_MULTI_EDGE(node)  (node->edge >= M)
#define IS_NODE_OF_TRIPLE_EDGE(node) (node->edge == 2 * M)

/* Checks whether this is the FIRST node (in the hash linked list) representing a double edge.
 * Used mainly for neighbors iterations in which each neighbor should be processed exactly once. */
#define IS_NEXT_NODE_THE_SAME(node) (node->next != NULL && node->next->id == node->id)

// Used when computing b_l(G) and b_d(G) for backward rejection.
#define NOT_IN_S(v)	  (v != S[0] && v != S[1] && v != S[2])
// Used when computing b_t(G) for backward rejection.
#define NOT_IN_S_3(v) (NOT_IN_S(v) && v != S[3])

// Points to the start and the end of the allocated adjacency nodes of a given vertex.
#define START_OF_POOL(v) (V[v].neighbors_pool)
#define END_OF_POOL(v)	 (V[v].next_available_node)

// The offsets in the points array of the single, double and triple edges sections.
#define START_OF_SINGLE_EDGES (2 * loop_count)
#define START_OF_DOUBLE_EDGES (M - 4 * double_count)
#define START_OF_TRIPLE_EDGES (M - 4 * double_count - 6 * triple_count)

/* Chooses a random loop or a random single/double/triple edge.
 * - RANDOM_LOOP points to the left cell of the loop edge in the points array.
 * - A random single/double/triple edge contains the edge orientation.		*/
#define RANDOM_LOOP			randInt(loop_count) * 2
#define RANDOM_SINGLE_EDGE 	randInt(START_OF_TRIPLE_EDGES - START_OF_SINGLE_EDGES) + START_OF_SINGLE_EDGES
#define RANDOM_DOUBLE_EDGE 	randInt(4 * double_count) + START_OF_DOUBLE_EDGES
#define RANDOM_TRIPLE_EDGE 	randInt(6 * triple_count) + START_OF_TRIPLE_EDGES

/* Points a randomly selected single edge to its left cell in the points array.
 * Sets the values for p1, p2 according to the original random orientation.  */
#define ADJUST_EDGE(p1,p2,edge){	\
	p1 = points[edge];				\
	if(IS_EVEN(edge)){				\
		p2 = points[edge + 1];		\
	}								\
	else{							\
		edge--;						\
		p2 = points[edge];			\
	}								\
}

// Creates a second adjacency node for a given endpoint of a double edge.
#define CREATE_SECOND_NODE(u, neighbor){ 					\
	addDoubleEdgeSecondNode(u, neighbor, END_OF_POOL(u));	\
	END_OF_POOL(u)++;										\
}

// Copies a pair to a new location in the points array. Updates new location if this is a single edge.
#define COPY_EDGE(src,dst){								\
	adjNode* node;										\
	GET_NODE(points[src], points[src + 1], node);		\
	if(node->edge < M){									\
		node->edge = dst;								\
		GET_NODE(points[src + 1], points[src], node);	\
		node->edge = dst;								\
	}													\
	points[dst]     = points[src];						\
	points[dst + 1] = points[src + 1];					\
}

/* While attempting to perform a Type I D-switch (including when reverted to INC-GEN),
 * checks for edges between the double edge and the two single edges in the given location.
 * Returns false (to NoDoubles) if a double edge is detected, otherwise indicates the
 * location of the detected edge, or indicates that more than one edge was detected.	 */
#define CHECK_FOR_EXISTING_EDGE(u,v){																\
	if(IS_EDGE(u,v)){																				\
		if(IS_DOUBLE_EDGE(u,v)){																	\
			LOG_DEBUG("Double edge (%d,%d) while attempting to perform Type I D-Switch.\n", u, v);	\
			return false;																			\
		}																							\
		existing_edge = (existing_edge == no_edge) ? u##_##v : multiple_edges;						\
	}																								\
}

// Indicates the type(s) of vertices relabeling required in a Type I Class B D-switch.
#define SWAP_SINGLE_EDGES   (existing_edge % 2 == 1)
#define REVERSE_ORIENTATION (existing_edge > 1)

// Used for determining the value of gamma (if an appropriate gamma exists).
#define EQ_24_LHS(gamma) 		(degree * n - 7 * FALL_2(degree) - 10 - 4 * gamma * (degree - 1) * (degree - 1))
#define DOES_EQ_25_HOLD(gamma) 	(n * FALL_2(degree - 1) > 															\
								2 * (1 + gamma) * (degree - 1) * FALL_2(degree) + 3 * degree * degree * (3 + degree))

// Printing formats for the statistics printout functions.
#define PERCENTAGE(a,b) 	 (b) > 0 ? 100.0 * (a) / (b) : 0
#define VAL_AND_PERCENT(a,b) a, PERCENTAGE(a, b)

/* Prints a summary of all attempts to generate a simple graph.
 * Invoked at the end of the run and after many consecutive failures. */
#define PRINT_OVERALL_STATS(mode){																					\
	LOG_##mode("\nOverall:\n"																						\
				"---------\n"																						\
				"Total attempts %d.\n"																				\
				"Pairing generation failures %d (%.1f%%), NoLoops failures %d (%.1f%%), "							\
				"NoTriples failures %d (%.1f%%), NoDoubles failures %d (%.1f%%), success %d (%.1f%%).\n\n",			\
				total_attempts, VAL_AND_PERCENT(generation_fails, total_attempts),									\
				VAL_AND_PERCENT(noLoops_fails,total_attempts), VAL_AND_PERCENT(noTriples_fails,total_attempts),		\
				VAL_AND_PERCENT(noDoubles_fails,total_attempts), VAL_AND_PERCENT(ii - 1, total_attempts));			\
}



/******************************************************************************************************************/
/**************************************************   FUNCTIONS   *************************************************/
/******************************************************************************************************************/

////////////////////////////////////////////////////////////////////////////////////////////////////////////
//LONGRAN generator (taken from Press et al)

#ifdef LONG_CYCLE_RNG

//#include <cmath>
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)
float ran2(int *idum) /* Adjusted: it's in range [0,1] */
{
	int j;
	long k;
	static long idum2=123456789;
	static long iy=0;
	static long iv[NTAB];
	float temp;

	if(*idum <= 0)
	{                            /* initialise  */
		if (-(*idum)<1) *idum=1; /* preventing idum=0 */
		else *idum = -(*idum);
		idum2 = (*idum);
		for (j=NTAB+7;j>=0;j--)
		{                        /*load the shuffle table (after 8 warm-ups) */
			k = (*idum)/IQ1;
			*idum = IA1*(*idum-k*IQ1)-k*IR1;
			if (*idum < 0) *idum += IM1;
			if (j< NTAB) iv[j] = *idum;
		}
		iy = iv[0];
	}

	k=(*idum)/IQ1;            /* start here when not initialising */
	*idum = IA1*(*idum - k*IQ1)-k*IR1; /* compute idum = (IA1*idum) %IM1 without   */
	if (*idum < 0) *idum += IM1;       /*   overflows by Schrage's method         */
	k = idum2/IQ2;
	idum2 = IA2*(idum2-k*IQ2) - k*IR2;  /* compute idum2 = (IA2*idum) %IM2 lokewise */
	if (idum2 < 0) idum2 += IM2;
	j = iy/NDIV;                       /* will be in range 0 .. NTAB - 1 */
	iy = iv[j] - idum2;                /* Here idum is shuffled, idum and idum2 are */
	iv[j] = *idum;                     /* combined to generate output */
	if(iy < 1) iy += IMM1;
	/*      if ((temp=AM*iy) > RNMX) return RNMX;   In original, this included as users don't expect endpoint value   */
	if ((temp=AM*iy) > 1) return 1.0;
	else  return temp;
}

int PRESSRAN(int n)
{
	double almost;
	almost = 1-1.0e-12;
	return floor(almost*ran2(&seed)*(n));
}

int LRAN(int n)
{
	double almost,rest,ndub;
	int rootn;
	ndub=n;
	almost = 1-1.0e-14;
	rootn=trunc( sqrt(ndub));
	rest=(ndub)/(rootn*1.0);
	return trunc(almost*(PRESSRAN(rootn)+ ran2(&seed))*rest );
}

#endif /* LONG_CYCLE_RNG */

// End of LONGRAN generator.
////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Frees allocated memory and closes output file(s) before terminating.
void finalize(exit_value val){

	// Free allocated memory (some pointers may be NULL pointers).
	free(V);
	free(points);
	free(hash_array);
	free(adjNode_pool);
	free(encountered);
	free(joint_neighbors_tracker);
	free(multi_edges);
	free(multi_edges_locations);
	free(rhoII);
	free(bA_approx);

	LOG_DEBUG("All memory freed.\n");

	// Close open files.
	fclose(log_file);
	if(output_file != stdout)
		fclose(output_file);

	exit(val);
}

/* Returns a random integer between 0 and (range - 1). Implementation depends on the RNG selected.
 * In case range is small enough and LONG_CYCLE_RNG is defined, a direct call to LRAN is made.
 * Otherwise, repeated calls to RAND_BASE_CALL are made, concatenating up to 60 random bits.
 * The largest possible multiple of range (with respect to the number of generated bits) is used
 * as a threshold - the generated number is used only if it is less than the threshold, and
 * otherwise rejected and a new number is drawn. This is done to avoid bias towards the smaller
 * values, and so, if RAND_BASE_CALL has a uniform distribution, so does randInt.  				*/
u64 randInt(u64 range){
#if SAFETY_CHECKS
	if(range > RAND_INT_MAX || range == 0){
		LOG_ERR("Error, call to randInt with %" PRIu64 ".\n"
				"randInt must be called with value 0 < k <= %" PRIu64 ".\n", range, RAND_INT_MAX);
		finalize(safety_check_failure);
	}
#endif

// If range is small enough a direct call to LRAN is sufficient.
#ifdef LONG_CYCLE_RNG
	if(range <= INT_MAX){
		return LRAN(range);
	}
#endif

	// Random bits concatenation is required.
	u64 num, threshold = 1;
	int iterations, ii;

#ifdef LONG_CYCLE_RNG
	// Exactly two iterations are required, each generates 30 bits.
	iterations = 2;
	threshold <<= 60;
#endif

#ifdef SIMPLE_RNG
	// Compute how many calls to rand() are required.
	iterations = 1;
	threshold <<= NUM_OF_RAND_BITS;
	while(range > threshold){
		threshold <<= NUM_OF_RAND_BITS;
		iterations += 1;
	}
#endif

	// Compute threshold.
	threshold -= (threshold % range);

	// Keep generating random numbers until a value smaller than threshold is obtained.
	while(true){
		num = RAND_BASE_CALL;
		for(ii = 1; ii < iterations; ii++){
			num = (num << NUM_OF_RAND_BITS) + RAND_BASE_CALL;
		}
		if(num < threshold)
			return (num % range);
	}
}

// Returns a random real number in the range [0,1). Resolution is 2^(-53).
double randReal(){
	return (1.0 * randInt(1ULL << 53)) / (1ULL << 53);
}

// Returns the multiplicity (0, 1, 2 or 3) of the edge (u,v).
int countEdgeMultiplicity(int u, int v){
	adjNode *node;

	GET_NODE(u,v,node);
	if(node == NULL)
		return 0;
	if(node->edge < M)
		return 1;
	if(node->edge == 2 * M)
		return 3;
	return 2;
}

// int-type comparison function given to qsort.
int intCompare(const void *a, const void *b){
	int val1 = *(const int*)a;
	int val2 = *(const int*)b;

	return val1 - val2;
}

// type_M comparison function given to qsort.
int type_M_Compare(const void *a, const void *b){
	type_M val1 = *(const type_M*)a;
	type_M val2 = *(const type_M*)b;

	if(val1 > val2)
		return 1;
	if(val1 == val2)
		return 0;
	return -1;
}

// Translates a numeric string to its value if it is in range (quits otherwise).
int strToInt(char* str){
	// Verify string does not begin with a zero.
	if(str[0] == '0'){
		LOG_ERR("ERROR: input parameters must be positive integers with no leading zeroes. Exiting.\n");
		finalize(input_failure);
	}

	// Verify value is in range.
	if(strlen(str) > 10 || atof(str) > S32_MAX){
		LOG_ERR("ERROR: input parameters must not exceed %d. Exiting.\n", S32_MAX);
		finalize(input_failure);
	}

	// Return the numeric value of the string (atoi is safe to use at this point).
	return atoi(str);
}

/* Asks for user's approval to continue when some problem is detected.
 * Quits the program if the user does not approve. 					*/
void approveContinuation(char* message){
	char buffer[100] = "";
	while(strcmp(buffer, "y") && strcmp(buffer,"n")){
		LOG_ERR("%s y/n\n", message);
		scanf("%100s", buffer);
	}
	if(!strcmp(buffer,"n"))
		finalize(user_termination);
}

// Prints the input format for the program.
void printInputFormat(){
	LOG_ERR("Program arguments must be in one of the following formats:\n"
			"1. <input_file>\n"
			"2. <input_file> <output_file>\n"
			"3. <input_file> <output_file> <log_file>\n"
			"or\n"
			"4. <number_of_graphs> <number_of_vertices> <degree>\n\n"

			"If an input file is used, format should match the one in option 4 (all white spaces are permitted).\n\n"

			"number_of_copies graphs will be generated, each d-regular on n vertices.\n\n"

			"- All parameters must be positive integers with no leading zeroes. \n"

			"- If an output file is not provided, generated graphs will be printed to stdout.\n"
			"- In order to avoid graphs output, set DO_NOT_OUTPUT to 1 in the code.\n"
			"- If a log file is provided, DEBUG_PRINTS should be set to 1 in the code.\n"
	);
}

// Checks whether the input for the program is via file (true) or via command line (false).
bool checkInputType(int argc, char** argv){
	// Command line input requires exactly 3 arguments (other than the program name).
	if(argc != 4)
		return true;

	// Command line input requires each argument to be a numeric string.
	for(int ii = 1; ii < 4; ii++){
		if(NOT_A_NUMBER(argv[ii]))
			return true;
	}

	// Input interpreted as command line input.
	return false;
}

// Extracts data from input file, quits if it is not in the correct format.
void readInputFile(char* input_file_name, char** input_array){

	// Open the input file.
	FILE *input_file = fopen(input_file_name, "r");
	if(!input_file){
		LOG_ERR("\nCould not open input file \"%s\". Exiting.\n\n", input_file_name);
		printInputFormat();
		finalize(file_failure);
	}

	// Attempt to read three numeric strings of length at most 10 from the file.
	for(int ii = 0; ii < 3; ii++){
		if((fscanf(input_file, "%11s", input_array[ii]) != 1) ||
			strlen(input_array[ii]) == 11 || NOT_A_NUMBER(input_array[ii])){
			fclose(input_file);
			LOG_ERR("\nWrong input file format!\n");
			printInputFormat();
			finalize(input_failure);
		}
	}

	// Verify EOF is reached.
	if(fscanf(input_file, "%*s") != EOF){
		fclose(input_file);
		LOG_ERR("\nWrong input file format!\n");
		printInputFormat();
		finalize(input_failure);
	}

	// Input file was valid and all data was read. Close file.
	fclose(input_file);
	LOG_DEBUG("Input file was read successfully.\n");
}

// Verifies the input is valid and no overflow occurs (quits otherwise), and computes M, M2, M3.
void verifyInput(char** input, int* number_of_copies){

	// Translate input numeric strings to their values.
	n 				  = strToInt(input[1]);
	degree 			  = strToInt(input[2]);
	*number_of_copies = strToInt(input[0]);

	// Verify input is graphical.
	if(!((IS_EVEN(n) || IS_EVEN(degree)) && degree < n)){
		LOG_ERR("ERROR: input is not graphical! Exiting.\n");
		finalize(input_failure);
	}

	// Verify degree is not too large.
	if(degree > (1 << 16)){
		LOG_ERR("Degree must not exceed %d.\n", 1 << 16);
		finalize(input_failure);
	}

	// Check for overflow if M is a 32-bit variable.
	if(!M_64 && S32_MAX < n * degree){
		LOG_ERR("Data overflow. M must not exceed %d. Set M_64 to 1 in the code.\n", S32_MAX);
		finalize(input_failure);
	}

	// Check for M3 overflow (this also ensures that M, M2 < RAND_INT_MAX).
	if(ULONG_LONG_MAX / n < FALL_3(degree)){
		LOG_ERR("Data overflow. M3 must not exceed %" PRIu64 ".\n", ULONG_LONG_MAX);
		finalize(input_failure);
	}

	// Valid input. Set M, M2, M3.
	M =  n * degree;
	M2 = n * FALL_2(degree);
	M3 = n * FALL_3(degree);
	LOG_DEBUG("Input is valid.\n");
}

// If fixed initial seed is used, verifies it has the correct sign according to the RNG.
void checkSeed(){
#if USE_FIXED_SEED
	if(floor(FIXED_SEED) != FIXED_SEED){
		LOG_ERR("\nFIXED_SEED must be an integer. Exiting.\n");
		finalize(configuration_failure);
	}

#ifdef SIMPLE_RNG
	if(FIXED_SEED < 0){
		LOG_ERR("\nSIMPLE_RNG is used with initial seed, but FIXED_SEED is negative (%d). Exiting.\n", FIXED_SEED);
		finalize(configuration_failure);
	}
#endif /* SIMPLE_RNG */

#ifdef LONG_CYCLE_RNG
	if(FIXED_SEED > 0){
		LOG_ERR("\nLONG_CYCLE_RNG is used with initial seed, but FIXED_SEED is positive (%d).\n"
				"Initialization part of the RNG will be skipped.\n\n", FIXED_SEED);
		approveContinuation("Continue without RNG initialization?");
	}
#endif /* LONG_CYCLE_RNG */

#endif /* USE_FIXED_SEED */
}

// Verify there is not any faulty user configuration.
void checkConfiguration(char** argv){
	// Verify that HASH_MULTIPLIER is in range.
	if(HASH_MULTIPLIER <= 0 || HASH_MULTIPLIER > S16_MAX){
		LOG_ERR("ERROR! HASH_MULTIPLIER must be a positive constant at most %d.\n", S16_MAX);
		finalize(configuration_failure);
	}

	// If fixed initial seed is used, verify it has the correct sign according to the RNG.
	checkSeed();

	// Check output configuration consistency.
	if(DO_NOT_OUTPUT && output_file != stdout){
		LOG_ERR("\nAn output file (%s) was provided but graph output is turned off.\n"
				"In order to enable graph output set DO_NOT_OUTPUT to 0 in the code.\n\n"
				"Proceeding without graph output.\n\n", argv[2]);
	}
}

// Compute all constants needed for the algorithm and revert to INC-GEN if necessary.
void computeConstants(){

	// If possible, set gamma to 0.9, otherwise set to 0.001, otherwise revert to INC-GEN.
	if((2.0 * FALL_3(degree - 1) * (degree - 1) * (degree - 1) < n * EQ_24_LHS(1)) && DOES_EQ_25_HOLD(0.9)){
		Gamma = 0.9;
	}
	else{
		if(DOES_EQ_25_HOLD(0.001))
			Gamma = 0.001;
		else
			revert_to_inc_gen = true;
	}

	// Compute epsilon and verify it is proper.
	if(!revert_to_inc_gen){
		epsilon = (4 * (1 + Gamma) * (degree - 1) * FALL_3(degree - 1)) / (degree * n * EQ_24_LHS(Gamma));
		if(epsilon <= 0 || epsilon >= 1){
			revert_to_inc_gen = true;
		}
	}

	// Cannot proceed with INC-REG.
	if(revert_to_inc_gen){
		LOG_REVERT("\nCould not find appropriate gamma and epsilon.\n\n");
		if(REVERT_AUTOMATICALLY)
			LOG_INFO("Reverting to INC-GEN.\n");
		else
			approveContinuation("Revert to INC-GEN?");

		// Compute B_L and B_D.
		if(22.0 * degree * degree * degree < M2){
			B_L = M2/M;
			B_D = floor((1.0 * M2 * M2) / (1.0 * M * M));
		}// else B_L = B_D = 0.

		// No need for other computations.
		return;
	}

	// Continuing with INC-REG.
	// Compute B_L, B_T and B_D.
	int log_dn = ceil(log(M));
	B_L = fmax(log_dn, 4 * degree);
	B_T = fmax(log_dn, (ceil(1.0 * degree * degree * degree) / n));
	B_D = floor(((1 + Gamma) * (degree - 1) * (degree - 1)) / 4);

	// Check if the lower bound used for loops removal b-rejection is small enough.
	if((8 * degree * B_D) + (degree * degree * B_L) >= M2 || (6 * degree * degree) - (4 * degree) >= M){
		LOG_REVERT("\nLoops b-rejection is assured.\n\n");
		if(REVERT_AUTOMATICALLY)
			LOG_INFO("Forbidding Loops.\n");
		else
			approveContinuation("Forbid loops (y) or abort (n) ?");

		// Forbid loops.
		B_L = 0;
	}

	// Check if the lower bound used for triple edges removal b-rejection is small enough.
	if(1.0 * degree * degree * (12 * B_D + 18 * B_T + 16 * degree + degree * degree) > M3 * (1 - 1.0e-15)){
		LOG_REVERT("\nTriple edges b-rejection probability is too high.\n\n");
		if(REVERT_AUTOMATICALLY)
			LOG_INFO("Forbidding triple edges.\n");
		else
			approveContinuation("Forbid triple edges (y) or abort (n) ?");

		// Forbid triple edges.
		B_T = 0;
	}

	// Check if the lower bound used for double edges removal b-rejection is small enough.
	if((degree - 1) * (16 * B_D + 3 * degree * degree + degree + 6) >= M2 ||
	   (4 * B_D * degree + 3 * degree * degree * (3 + degree)) >= M2){
		LOG_REVERT("\nDouble edges b-rejection is assured.\n\n");
		if(REVERT_AUTOMATICALLY)
			LOG_INFO("Forbidding double edges.\n");
		else
			approveContinuation("Forbid double edges (y) or abort (n) ?");

		// Forbid triple edges.
		B_D = 0;

		// Remaining computations are not needed as they are only used for double edges removal.
		return;
	}

	// Compute the probabilities rhoII(j) and the lower bounds bA(j) for 0 <= j <= B_D.
	ALLOCATE_MEMORY(bA_approx, double, B_D + 1);
	ALLOCATE_MEMORY(rhoII,	   double, B_D + 1);

	int ii;
	double mI, mII, mA, mB, x_curr, x_prev, mI_prev;

	double rhoI = 1 - epsilon;					// Lemma 7
	rhoII[0] = rhoII[B_D] = 0.0;				// Eq. (16)

	// bA_approx = mA below when computing rhoII, and is used in NoDoubles if no precise computation is applicable.
	for (ii = 0; ii <= B_D; ii++){
		bA_approx[ii] = 1.0 * M2 * (M2 - (degree - 1) * (16 * ii + 3 * degree * degree + degree + 6));		// Eq. (21)
	}

	// In order to compute x_{B_D} according to (14a), must compute mI and mB for B_D.
	mI_prev = 4.0 * B_D * (M - 4 * B_D) * (M - 4 * B_D);													// Eq. (19) for B_D
	mB   = 16.0 * B_D * (M3 - degree * (8 * B_D * degree + 3 * degree * degree * (3 + degree)));			// Eq. (22) for B_D
	x_prev = mI_prev / (mI_prev - rhoI * mB);																// Eq. (14a)

	// Compute rhoII(j) for 0 < j < B_D.
	for (ii = B_D - 1; ii > 0; ii--){
		mI  = 4.0  * ii * (M - 4 * ii) * (M - 4 * ii);														// Eq. (19)
		mII = 16.0 * ii * (degree - 1) * (degree - 1) * (degree - 2) * (degree - 3);						// Eq. (20)
		mA  = bA_approx[ii];																				// Eq. (21)
		mB  = 16.0 * ii * (M3 - degree * (8 * ii * degree + 3 * degree * degree * (3 + degree)));			// Eq. (22)

		// Compute x_j and rhoII(j).
		x_curr	  = mI * (mI_prev + x_prev * rhoI * mA) / (mI_prev * (mI - rhoI * mB));						// Eq. (13)
		rhoII[ii] = rhoI * x_prev * mII / (x_curr * mI_prev);												// Eq. (15)

		// Prepare for next iteration.
		x_prev	= x_curr;
		mI_prev = mI;
	}

	// Checking if bA requires approximation in NoDoubles or can be computed precisely, using the fact that bA < M2 * M2.
	if(RAND_INT_MAX / M2 >= M2){
		// No need for approximation.
		free(bA_approx);
		bA_approx = NULL;
	}
}

// Initializes the RNG according to its definitions.
void initializeRNG(){
#ifdef LONG_CYCLE_RNG
#if USE_FIXED_SEED
	seed = FIXED_SEED;
#else
	// Use random initial seed.
	srand(time(NULL));
	seed = -rand();
#endif /* USE_FIXED_SEED */
#endif /* LONG_CYCLE_RNG */

#ifdef SIMPLE_RNG
#if USE_FIXED_SEED
	srand(FIXED_SEED);
#else
	// Use random initial seed.
	srand(time(NULL));
#endif /* USE_FIXED_SEED */
#endif /* SIMPLE_RNG */
}

// Allocates all dynamic memory and initializes all the data that depends on the degree sequence alone.
void initialize(){
	int ii, jj;
	u64 sum_deg;

	// Initialize the random number generator.
	initializeRNG();

	// Compute all constants needed for ING-REG and revert to INC-GEN if necessary.
	computeConstants();

	// Compute hash related values.
	hash_cells_per_vertex = ceil(degree * (HASH_MULTIPLIER));
	hash_cell_size = ceil((1.0 * n) / hash_cells_per_vertex);

	// Allocate all dynamic memory.
	ALLOCATE_MEMORY(V,		 		vertex,	 	n);
	ALLOCATE_MEMORY(points, 		int,	 	M);
	ALLOCATE_MEMORY(adjNode_pool,	adjNode, 	M);
	ALLOCATE_MEMORY(hash_array, 	adjNode*, 	n * hash_cells_per_vertex);
	ALLOCATE_MEMORY(encountered, 	int,		2 * degree * degree);
	ALLOCATE_MEMORY(multi_edges, 	multiEdge,  B_D + B_T);
	ALLOCATE_MEMORY(multi_edges_locations, type_M, 2 * B_D + 3 * B_T);

	// joint_neighbors_tracker is the only one needed to be set to 0.
	ALLOCATE_INITIALIZED_MEMORY(joint_neighbors_tracker, jointNeighbors, n);

	/* Create the points array and initialize "has_loop", "hash" and "neighbors_pool" fields of each v_i.
	 * "single_edges_num" and "next_available_node" are initialized before each pairing is created. 	*/
	for(ii = 0, sum_deg = 0; ii < n; ii++){
		V[ii].has_loop		 = false;
		V[ii].hash			 = hash_array   + ii * hash_cells_per_vertex;
		V[ii].neighbors_pool = adjNode_pool + ii * degree;

		// Points corresponding to v_i.
		for(jj = 0; jj < degree; jj++){
			points[sum_deg] = ii;
			sum_deg++;
		}
	}

	LOG_DEBUG("\nInitialization completed successfully.\n"
				"n = %d, degree = %d, M = %" PRIu64 ", M2 = %"PRIu64", M3 = %"PRIu64", B_L = %d, B_D = %d",
				n, (int) degree, M, M2, M3, B_L, B_D);
	if(!revert_to_inc_gen)
		LOG_DEBUG(" B_T = %d, gamma = %.3f, epsilon ~ %f", B_T, Gamma, epsilon);
	LOG_DEBUG(".\n");
}

// Prints generated pairing related statistics at the end of the run.
void printStatisticsPairings(int generation_fails, int total_attempts){
	if(generation_fails == 0){
		LOG_INFO("\n----- All pairing generations were successful.\n");
		return;
	}

	LOG_INFO("\nPairing generations:\n"
			 "---------------------\n"
			 "%d successful, %d failed. Failures according to the first detected problem:\n",
			 total_attempts - generation_fails, generation_fails);

	if(revert_to_inc_gen)
		LOG_INFO("double loop %d, triple edge %d, too many loops %d, too many double edges %d.\n",
				 double_loop, many_triple_edges, many_loops, many_multiple_edges);

	else
		LOG_INFO("double loop %d, quadruple edge %d, too many loops %d, too many double edges %d, "
				 "too many triple edges %d, too many multiple* edges %d.\n"
				 "(* edges with multiplicity > 1, exact multiplicity undetermined upon rejection)\n",
				 double_loop, quadruple_edge, many_loops, many_double_edges,
				 many_triple_edges, many_multiple_edges);
}

// Print loops related statistics at the end of the run.
void printStatisticsLoops(){
	if(total_L_switch_calls == 0){
		if(B_L == 0)
			LOG_INFO("\n----- No loops were allowed.\n");
		else
			LOG_INFO("\n----- No calls to NoLoops were made.\n");

		return;
	}

	int total_rejections = L_f_rej + L_b_rej;
	int success = total_L_switch_calls - total_rejections;
	LOG_INFO("\nLoops:\n"
			 "-------\n"
			 "Total calls to NoLoops %d, average loops per call %.1f, total rejections %d (%.1f%%).\n"
			 "Total loop removal attempts %d, Forward rejections %d (%.1f%%), Backward rejections %d (%.1f%%), "
			 "success %d (%.1f%%).\n",
			 total_noloops, 1.0 * total_loop_num / total_noloops, VAL_AND_PERCENT(total_rejections ,total_noloops),
			 total_L_switch_calls, VAL_AND_PERCENT(L_f_rej ,total_L_switch_calls),
			 VAL_AND_PERCENT(L_b_rej, total_L_switch_calls), VAL_AND_PERCENT(success, total_L_switch_calls));
}

// Print triple edges related statistics at the end of the run.
void printStatisticsTriples(){
	if(total_T_switch_calls == 0){
		if(B_T == 0)
			LOG_INFO("\n----- No triple edges were allowed.\n");
		else
			LOG_INFO("\n----- No calls to NoTriples were made.\n");

		return;
	}

	int total_rejections = T_f_rej + T_b_rej;
	int success = total_T_switch_calls - total_rejections;
	LOG_INFO("\nTriple edges:\n"
			 "--------------\n"
			 "Total calls to NoTriples %d, average triple edges per call %.1f, total rejections %d (%.1f%%).\n"
			 "Total triple edge removal attempts %d, Forward rejections %d (%.1f%%), "
			 "Backward rejections %d (%.1f%%), success %d (%.1f%%).\n",
			 total_notriples, 1.0 * total_triple_edges_num / total_notriples,
		 	 VAL_AND_PERCENT(total_rejections ,total_notriples),
			 total_T_switch_calls, VAL_AND_PERCENT(T_f_rej ,total_T_switch_calls),
			 VAL_AND_PERCENT(T_b_rej, total_T_switch_calls), VAL_AND_PERCENT(success, total_T_switch_calls));
}

// Print double edges related statistics at the end of the run.
void printStatisticsDoubles(){
	if(total_D_switch_calls == 0){
		if(B_D == 0)
			LOG_INFO("\n----- No double edges were allowed.\n");
		else
			LOG_INFO("\n----- No calls to NoDoubles were made.\n");

		return;
	}

	int total_rejections, success;

	if(revert_to_inc_gen){
		total_rejections = D_f_rej + D_b_rej;
		success 		 = total_D_switch_calls - total_rejections;
		LOG_INFO("\nDouble edges:\n"
				 "--------------\n"
				 "Total calls to NoDoubles %d, average double edges per call %.1f, total rejections %d (%.1f%%).\n"
				 "Total double edge removal attempts %d, Forward rejections %d (%.1f%%), "
				 "Backward rejections %d (%.1f%%), success %d (%.1f%%).\n",
				 total_nodoubles, 1.0 * total_double_edges_num / total_nodoubles,
				 VAL_AND_PERCENT(total_rejections ,total_nodoubles),
				 total_D_switch_calls, VAL_AND_PERCENT(D_f_rej ,total_D_switch_calls),
				 VAL_AND_PERCENT(D_b_rej, total_D_switch_calls), VAL_AND_PERCENT(success, total_D_switch_calls));

		return;
	}

	D_f_rej = type_I_f_rej  + type_II_f_rej;
	D_b_rej = Class_A_b_rej + Class_B_b_rej;

	total_rejections = D_f_rej + D_b_rej + type_rej;
	success = total_D_switch_calls - total_rejections;

	int Type_II_Class_B = total_Type_II - type_II_f_rej;
	int total_Class_B   = total_Type_I_Class_B + Type_II_Class_B;

	LOG_INFO("\nDouble edges:\n"
			 "--------------\n"
			 "GENERAL\n"
			 "Total calls to NoDoubles %d, average double edges per call %.1f, total rejections %d (%.1f%%).\n"
			 "Total double edge removal attempts %d, Forward rejections %d (%.1f%%), Backward rejections %d (%.1f%%), "
			 "type rejection %d (%.1f%%), success %d (%.1f%%).\n\n"
			 "TYPES AND CLASSES\n"
			 "Type selection:      Type  I %d (%.1f%%), Type II %d (%.1f%%), type rejection %d (%.1f%%).\n"
			 "Type I calls:        Class A %d (%.1f%%), Class B %d (%.1f%%), Forward rejection %d (%.1f%%).\n"
			 "Type II calls:       Class B %d (%.1f%%), Forward rejection %d (%.1f%%).\n"
			 "Backwards rejection: Class A %d of %d (%.1f%%), Class B %d of %d (%.1f%%).\n",
			 total_nodoubles, 1.0 * total_double_edges_num / total_nodoubles,
			 VAL_AND_PERCENT(total_rejections ,total_nodoubles),
			 total_D_switch_calls, VAL_AND_PERCENT(D_f_rej ,total_D_switch_calls),
			 VAL_AND_PERCENT(D_b_rej, total_D_switch_calls), VAL_AND_PERCENT(type_rej, total_D_switch_calls),
			 VAL_AND_PERCENT(success, total_D_switch_calls),
			 VAL_AND_PERCENT(total_Type_I, total_D_switch_calls), VAL_AND_PERCENT(total_Type_II, total_D_switch_calls),
			 VAL_AND_PERCENT(type_rej, total_D_switch_calls),
			 VAL_AND_PERCENT(total_Class_A, total_Type_I), VAL_AND_PERCENT(total_Type_I_Class_B, total_Type_I),
			 VAL_AND_PERCENT(type_I_f_rej, total_Type_I),
			 VAL_AND_PERCENT(Type_II_Class_B, total_Type_II), VAL_AND_PERCENT(type_II_f_rej, total_Type_II),
			 Class_A_b_rej, total_Class_A, PERCENTAGE(Class_A_b_rej, total_Class_A),
			 Class_B_b_rej, total_Class_B, PERCENTAGE(Class_B_b_rej, total_Class_B));
}

/* Prints all pairs in the points array according to their multiplicity
 * and by their order: loops, single edges, triple edges, double edges. */
void _printPairing(){
	for(u64 ii = 0; ii < M/2; ii++)
		LOG_DEBUG("pair %" PRIu64": %d - %d\n", ii, points[2 * ii], points[2 * ii + 1]);
}

// Prints the "edge" field for all allocated adjacency nodes, vertex by vertex.
void printEdgeLocations(){
	adjNode *node;
	for(int ii = 0; ii < n; ii++){
		for(node = START_OF_POOL(ii); node < END_OF_POOL(ii) - 1; node++){
			LOG_DEBUG("(%d,%d) --> %" PRINT_M", ", ii, node->id, node->edge);
		}
		LOG_DEBUG("(%d,%d) --> %" PRINT_M".\n", ii, node->id, node->edge);
	}
}

// Prints all neighbors of a given vertex by iterating its neighbors pool.
void printNeighborsByPool(int u){
	adjNode *node;
	LOG_DEBUG("Neighbors of vertex %d: ",u);
	for(node = START_OF_POOL(u); node < END_OF_POOL(u) - 1; node++){
		LOG_DEBUG("%d, ", node->id);}
	LOG_DEBUG("%d", node->id);
	if(V[u].has_loop){
		LOG_DEBUG(", and also has a loop.\n");}
	else{
		LOG_DEBUG(".\n");}
}

// Prints all neighbors of a given vertex by iterating its hash.
void _printNeighborsByHash(int u){
	adjNode* node;
	LOG_DEBUG("Neighbors of vertex %d (single edges = %d): ", u, V[u].single_edges_num);
	for(int ii = 0; ii < hash_cells_per_vertex; ii++){
		node = V[u].hash[ii];
		LOG_DEBUG("[");
		if(node == NULL) LOG_DEBUG("*");
		while(node != NULL){
			LOG_DEBUG("%d", node->id);
			if(node->edge >= M){
				if(node->edge == 2 * M)
					LOG_DEBUG("(T)");
				else
					LOG_DEBUG("(D)");
			}
			node = node->next;
			if(node != NULL)
				LOG_DEBUG(",");
		}
		LOG_DEBUG("]");
	}
	if(V[u].has_loop)
		LOG_DEBUG(" and also has a loop.\n");
	else
		LOG_DEBUG(".\n");
}

// Iterates all vertices and for each prints all its neighbors.
void _printAllNeighbors(){
/*
	LOG_DEBUG("Print by pool:\n");
	for(int ii = 0; ii < n; ii++)
		printNeighborsByPool(ii);
*/

	LOG_DEBUG("Print by hash:\n");
	for(int ii = 0; ii < n; ii++)
		_printNeighborsByHash(ii);
}

// Outputs all edges in the graph by their order in the points array.
void outputGraphUnsortedEdges(){
	char buffer[100];

	// Set the column width.
	int width = COLUMN_WIDTH;
	if(width == 0)
		width = log(n) + 8;

	for(u64 ii = 0; ii < M/2; ii++){
		if(ii % EDGES_PER_LINE == 0)
			OUTPUT("\n");
		sprintf(buffer, "(%d,%d)", points[2 * ii], points[2 * ii + 1]);
		OUTPUT("%-*s", width, buffer);
	}
	OUTPUT("\n\n\n");
}

// Outputs the graph as unordered lists of neighbors per vertex.
void outputGraphUnsortedNeighbors(){
	adjNode *node;

	// Iterate all vertices.
	for(int ii = 0; ii < n; ii++){
#if PRINT_EMPTY_LISTS
		OUTPUT("\n%d: ", ii);
		node = START_OF_POOL(ii);
#else
		// Look for the first neighbor.
		for(node = START_OF_POOL(ii); node < END_OF_POOL(ii); node++){
			if(node->id > ii){
				OUTPUT("\n%d: %d ", ii, node->id);
				node++;
				break;
			}
		}
#endif

		// Iterate remaining/all neighbors of the current vertex.
		for(; node < END_OF_POOL(ii); node++){
			if(node->id > ii){
				OUTPUT("%d ", node->id);
			}
		}
	}

	OUTPUT("\n\n\n");
}

/* Outputs the graph in a sorted way, according to user configuration.
 * If OUTPUT_GRAPH_BY_VERTEX is on, prints for each vertex its ordered
 * 									list of neighbors of larger index.
 * Otherwise, prints all edges in the graph in lexicographical order. */
void outputGraphSorted(){
	int 	ii, jj, cell_length = 0, vertex_length = 0;
	int 	*list_to_array;
	adjNode *node, **head;

	ALLOCATE_MEMORY(list_to_array, int, degree);

#if !OUTPUT_GRAPH_BY_VERTEX
	int  total_edges_printed = 0;
	char buffer[100];

	// Set the column width.
	int width = COLUMN_WIDTH;
	if(width == 0)
		width = log(n) + 8;
#endif

	// Iterate all vertices.
	for(ii = 0; ii < n - 1; ii++){
		/* For each vertex, iterate its hash table. There is order between cells, but not within each cell.
		 * Skip all cells corresponding to vertices of smaller index than ii. */
		for(head = GET_CELL(ii, ii + 1); head < V[ii + 1].hash; head++){
			// Sort every non-empty hash cell.
			if(*head != NULL){
				node = *head;
				while(node != NULL){
					list_to_array[vertex_length + cell_length] = node->id;
					cell_length++;
					node = node->next;
				}

				qsort(list_to_array + vertex_length, cell_length, sizeof(int), intCompare);
				vertex_length += cell_length;
				cell_length = 0;
			}
		}

		jj = 0;
		// Skip neighbors with index smaller than ii (may appear only in the first iterated cell of the vertex).
		while((jj < vertex_length) && (list_to_array[jj] < ii))
			jj++;

#if OUTPUT_GRAPH_BY_VERTEX
		if(jj < vertex_length || PRINT_EMPTY_LISTS)
			OUTPUT("\n%d: ", ii);
		while(jj < vertex_length){
			OUTPUT("%d ", list_to_array[jj]);
			jj++;
		}
#else
		// Print all other edges.
		while(jj < vertex_length){
			if(total_edges_printed % EDGES_PER_LINE == 0)
				OUTPUT("\n");
			total_edges_printed++;
			sprintf(buffer, "(%d,%d)", ii, list_to_array[jj]);
			OUTPUT("%-*s", width, buffer);
			jj++;
		}
#endif

		vertex_length = 0;
	}

#if OUTPUT_GRAPH_BY_VERTEX
	// The last vertex has no neighbors of larger index.
	if(PRINT_EMPTY_LISTS)
		OUTPUT("\n%d:", n-1);
#endif

	free(list_to_array);
	OUTPUT("\n\n\n");
}

// Verifies each vertex appears the correct number of times in the points array (quits otherwise).
void verifyPointsCount(){
	u64	ii;
	int	*count;

	ALLOCATE_INITIALIZED_MEMORY(count, int, n);

	for(ii = 0; ii < M; ii++)
		count[points[ii]]++;

	for(ii = 0; ii < n; ii++){
		if(degree != count[ii]){
			LOG_ERR("ERROR: vertex %" PRIu64 " has degree %d but appears %d times.\n",
					ii, (int) degree, count[ii]);
			_printPairing();
			free(count);
			finalize(safety_check_failure);
		}
	}

	free(count);
	LOG_DEBUG("pairing count verified.\n");
}

// Verifies the pairs in the non-single sections are valid (quits otherwise).
void verifyPointsSections(){
	u64 ii;

	// Check loops section.
	for(ii = 0; ii < START_OF_SINGLE_EDGES; ii += 2){
		if(points[ii] != points[ii + 1]){
			LOG_ERR("ERROR: pair %" PRIu64 " should be a loop but is not.\n", ii / 2);
			_printPairing();
			finalize(safety_check_failure);
		}
	}

	// Check triple edges section.
	for(ii = START_OF_TRIPLE_EDGES; ii < START_OF_DOUBLE_EDGES; ii += 6){
		if(!(ARE_PAIRS_EQUAL(points[ii], points[ii + 1], points[ii + 2], points[ii + 3]) &&
			 ARE_PAIRS_EQUAL(points[ii], points[ii + 1], points[ii + 4], points[ii + 5]))){
			LOG_ERR("ERROR: pairs %" PRIu64 " - %" PRIu64 " - %" PRIu64
					" should represent a triple edge but are not.\n", ii / 2, ii / 2 + 1, ii / 2 + 2);
			_printPairing();
			finalize(safety_check_failure);
		}
	}

	// Check double edges section.
	for(ii = START_OF_DOUBLE_EDGES; ii < M; ii += 4){
		if(!ARE_PAIRS_EQUAL(points[ii], points[ii + 1], points[ii + 2], points[ii + 3])){
			LOG_ERR("ERROR: pairs %" PRIu64 " - %" PRIu64
					" should represent a double edge but are not.\n", ii / 2, ii / 2 + 1);
			_printPairing();
			finalize(safety_check_failure);
		}
	}

	LOG_DEBUG("pairing sections verified.\n");
}

// Verifies the points array is proper (quits otherwise).
void _checkPointsArray(){
	// Verify each vertex appears the correct number of times in the points array.
	verifyPointsCount();

	// Verify the pairs in the non-single sections are valid.
	verifyPointsSections();
}

// Verifies the correctness of the "edge" field of every adjacency node (quits otherwise).
void verifyEdgeIndicators(){
	u64		jj;
	adjNode *node;

	for(int ii = 0; ii < n; ii++){
		for(node = START_OF_POOL(ii); node < END_OF_POOL(ii); node++){
			// Verify "edge" field is in range.
			if((node->edge < 0) || (node->edge > 2 * M)){
				LOG_ERR("ERROR: edge %d->%d indicator wrong (marked %" PRINT_M").\n",
						ii, node->id, node->edge);
				finalize(safety_check_failure);
			}

			if(IS_NODE_OF_SINGLE_EDGE(node)){
				// Verify the edge appears in the indicated location in the points array.
				if(!ARE_PAIRS_EQUAL(ii, node->id, points[node->edge], points[node->edge + 1])){
					LOG_ERR("ERROR: edge %d->%d indicator wrong (marked single).\n", ii, node->id);
					_printNeighborsByHash(ii);
					printEdgeLocations();
					finalize(safety_check_failure);
				}
			}
			else{
				if(IS_NODE_OF_TRIPLE_EDGE(node)){
					// Verify the edge appears in the triple edges section in the points array.
					for(jj = START_OF_TRIPLE_EDGES; jj < START_OF_DOUBLE_EDGES; jj += 6){
						if(ARE_PAIRS_EQUAL(ii, node->id, points[jj], points[jj + 1]))
							break;
					}
					if(jj >= START_OF_DOUBLE_EDGES){
						LOG_ERR("ERROR: edge %d->%d indicator wrong (marked triple).\n", ii, node->id);
						_printNeighborsByHash(ii);
						printEdgeLocations();
						finalize(safety_check_failure);
					}
				}
				else{
					// Verify the edge appears in the double edges section in the points array.
					for(jj = START_OF_DOUBLE_EDGES; jj < M; jj += 4){
						if(ARE_PAIRS_EQUAL(ii, node->id, points[jj], points[jj + 1]))
							break;
					}
					if(jj >= M){
						LOG_ERR("ERROR: edge %d->%d indicator wrong (marked double).\n", ii, node->id);
						_printNeighborsByHash(ii);
						printEdgeLocations();
						finalize(safety_check_failure);
					}
				}
			}
		}
	}

	LOG_DEBUG("Edge indicators verified.\n");
}

// Verifies the the correctness and consistency of the different data structures (quits otherwise).
void _verifyDataStructures(){

	u64		ii;
	adjNode *node;
	bool 	faulty_database = false;
	int 	singles, doubles, triples;
	int 	*multi_edge_count;

	// Verifies the points array is proper.
	_checkPointsArray();

	// Verify the correctness of the "edge" field of every adjacency node.
	verifyEdgeIndicators();

	// Verify for each vertex that the loops, single edges, and multiple edges incident with it sum up to its degree.
	ALLOCATE_INITIALIZED_MEMORY(multi_edge_count, int, n);
	for(ii = START_OF_TRIPLE_EDGES; ii < M; ii++){
		multi_edge_count[points[ii]]++;
	}
	for(ii = 0; ii < n; ii++){
		if(degree != V[ii].single_edges_num + 2 * V[ii].has_loop + multi_edge_count[ii]){
			LOG_ERR("ERROR: vertex %" PRIu64 " has degree %d but single edges number is %d, "
					"multiple edge count is %d, and it has %s loop.\n",
					ii, (int) degree, V[ii].single_edges_num, multi_edge_count[ii],
					V[ii].has_loop? "a" : "no");
			_printNeighborsByHash(ii);
			faulty_database = true;
		}

		singles = doubles = triples = 0;
		for(node = START_OF_POOL(ii); node < END_OF_POOL(ii); node++){
			if(node->edge < M){
				singles++;
			}
			else if(node->edge == 2 * M)
				triples++;
			else if(!IS_NEXT_NODE_THE_SAME(node))
				doubles++;
		}
		if(singles != V[ii].single_edges_num){
			LOG_ERR("ERROR: vertex %" PRIu64 " has %d single nodes but single edges number is %d.\n",
					ii, singles, V[ii].single_edges_num);
			_printNeighborsByHash(ii);
			faulty_database = true;
		}
		if(2 * doubles + 3 * triples != multi_edge_count[ii]){
			LOG_ERR("ERROR: vertex %" PRIu64 " has %d double nodes and %d triple nodes but multi count is %d.\n",
					ii, doubles, triples, multi_edge_count[ii]);
			_printNeighborsByHash(ii);
			faulty_database = true;
		}
	}

	free(multi_edge_count);

	if(faulty_database){
		finalize(safety_check_failure);
	}

	// Verify that P2 holds the correct value.
	u64 tmp_P2 = 0;
	for(ii = 0; ii < n; ii++){
		if(!V[ii].has_loop)
			tmp_P2 += FALL_2(V[ii].single_edges_num);
	}

	if(P2 != tmp_P2){
		LOG_ERR("ERROR: P2 %"PRIu64" != tmp_P2 %"PRIu64".\n", P2, tmp_P2);
		finalize(safety_check_failure);
	}

	LOG_DEBUG("Data structures verified.\n");
}

// Adds a new single neighbor to a vertex u.
void addNewSingleNeighbor(int u, int neighbor, type_M location){
#if SAFETY_CHECKS
	if(END_OF_POOL(u) == START_OF_POOL(u) + degree){
		LOG_ERR("ERROR: vertex %d has already allocated all its nodes!\n", u);
		finalize(safety_check_failure);
	}
	if(IS_EDGE(u, neighbor)){
		LOG_ERR("ERROR: vertex %d is already a neighbor of %d!\n", neighbor, u);
		finalize(safety_check_failure);
	}
#endif

	// Allocate a new adjacency node.
	adjNode *node = END_OF_POOL(u);
	END_OF_POOL(u)++;

	// Update the node's fields.
	node->id   = neighbor;
	node->edge = location;

	// Insert node to hash.
	HEAD_INSERT(u, node);
}

// Inserts to the hash a second adjacency node of a double neighbor. The two nodes are kept consecutively.
void addDoubleEdgeSecondNode(int u, int neighbor, adjNode *new_node){

	adjNode *old_node;

	// Find the existing adjacency node in the hash.
	GET_NODE(u, neighbor, old_node);

#if SAFETY_CHECKS
	if(old_node == NULL){
		LOG_ERR("ERROR: vertex %d is not an existing neighbor of %d!\n", neighbor, u);
		finalize(safety_check_failure);
	}
#endif

	// Insert the new node to the list immediately after the existing one.
	new_node->next = old_node->next;
	old_node->next = new_node;

	// Set id for the new node and mark both nodes as belonging to a double edge.
	new_node->id   = neighbor;
	old_node->edge = M;
	new_node->edge = M;
}

// Changes one of the neighbors of a vertex u to a new, single, neighbor.
void changeNeighbor(int u, int old_neighbor, int new_neighbor, type_M location){

	adjNode* node;
	LOG_DEBUG("Adding %d as a single neighbor of %d instead of %d.\n", new_neighbor, u, old_neighbor);

#if SAFETY_CHECKS
	if(IS_NO_EDGE(u, old_neighbor)){
		LOG_ERR("ERROR: vertex %d is not a neighbor of %d!\n", old_neighbor, u);
		finalize(safety_check_failure);
	}
	if(IS_EDGE(u, new_neighbor)){
		LOG_ERR("ERROR: vertex %d is already a neighbor of %d!\n", new_neighbor, u);
		finalize(safety_check_failure);
	}
#endif

	// If the two vertices belong to the same hash cell, just update the content of the node.
	if(GET_CELL(u, old_neighbor) == GET_CELL(u, new_neighbor)){
		GET_NODE(u, old_neighbor, node);
		node->id   = new_neighbor;
		node->edge = location;
		return;
	}

	// The vertices belong to different hash cells. First extract node from old cell.
	EXTRACT_NODE(u, old_neighbor, node);

	// Update node's fields.
	node->id   = new_neighbor;
	node->edge = location;

	// Insert the new neighbor to the top of its hash cell.
	HEAD_INSERT(u, node);
}

/* Changes one of the neighbors of a vertex u to another, existing, neighbor, creating
 * a double edge. This happens only when a Class B D-switch is performed.			*/
void changeNeighborToDouble(int u, int old_neighbor, int new_neighbor){

	adjNode* node;
	LOG_DEBUG("Adding %d as a double neighbor of %d instead of %d.\n", new_neighbor, u, old_neighbor);

#if SAFETY_CHECKS
	if(IS_NO_EDGE(u, old_neighbor)){
		LOG_ERR("ERROR: vertex %d is not a neighbor of %d!\n", old_neighbor, u);
		finalize(safety_check_failure);
	}
	if(!IS_SINGLE_EDGE(u, new_neighbor)){
		LOG_ERR("ERROR: vertex %d is not a single neighbor of %d!\n", new_neighbor, u);
		finalize(safety_check_failure);
	}
#endif

	// First extract node of old neighbor.
	EXTRACT_NODE(u, old_neighbor, node);

	// Insert the new neighbor after the old node of that vertex.
	addDoubleEdgeSecondNode(u, new_neighbor, node);
}

/* For a double edge (v1,v2), translates an index in [0 .. d-3] to a unique
 * adjacency node of v1 not corresponding to v2. Used for Class B D-switches. */
adjNode* indexToNode(int v1, int v2, int index){
	// First locate the node pointed by the index.
	adjNode* node = START_OF_POOL(v1) + index;

	// If the index corresponds to a neighbor other than v2 return it.
	if(node->id != v2)
		return node;

	/* Adjustment is needed. One of the last two nodes in the pool is returned
	 * according to the index and the location of the two nodes representing v2. */
	if(IS_NEXT_NODE_THE_SAME(node)){
		node = END_OF_POOL(v1) - 1;
		if(node->id == v2)
			node--;
	}
	else{
		node = END_OF_POOL(v1) - 2;
		if(node->id == v2)
			node++;
	}
	return node;
}

// Chooses two distinct random edges intersecting the double edge (v1,v2) at v1.
void selectRandomNeighborsOfDoubleEdge(int v1, int v2, adjNode** v3, adjNode** v4){
	int index1, index2, rand_num;

	// Choose u.a.r. two ordered distinct integers in the range [0 .. d-3].
	rand_num = randInt(FALL_2(degree - 2));
	index1 = rand_num % (degree  - 2); // Uniform distribution of [0 .. d-3].
	index2 = rand_num / (degree  - 2); // Uniform distribution of [0 .. d-4].
	if(index2 == index1)
		index2 = degree - 3;

	// Translate these integers to nodes of neighbors of v1 other than v2.
	*v3 = indexToNode(v1, v2, index1);
	*v4 = indexToNode(v1, v2, index2);
}

// Chooses a random edge intersecting the single edge (v1,v2) at v1.
adjNode* selectRandomNeighborsOfSingleEdge(int v1, int v2){
	int index = randInt(degree - 1);
	// Translate the index to a neighbor of v1 other than v2.
	adjNode* node = START_OF_POOL(v1) + index;
	if(node->id == v2)
		node = END_OF_POOL(v1) - 1;

	return node;
}

// Reject (i.e. return true) with probability 1 - numerator/denominator.
bool checkBackwardsRejectionNonIncremental(u64 numerator, u64 denominator){
	return(randInt(denominator) >= numerator);
}

// Reject (i.e. return true) with probability 1 - (num1*num2)/(denom1*denom2).
bool checkBackwardsRejectionIncremental(u64 num1, u64 num2, u64 denom1, u64 denom2){
	if(RAND_INT_MAX / denom1 >= denom2){
		// Generate a random number just once if possible.
		return(checkBackwardsRejectionNonIncremental(num1 * num2, denom1 * denom2));
	}
	else{
		// Compute incrementally.
		return(randInt(denom1) >= num1 || randInt(denom2) >= num2);
	}
}

/* Generates a random pairing of all the points and creates all relevant data structures accordingly.
 * All pairs corresponding to loops appear at the beginning of the points array, followed by pairs
 * corresponding to single edges, triple edges (three consecutive copies of each such pair), and finally
 * double edges (two consecutive copies of each such pair).
 * Returns false if one of the following occurs:
 * 		a double loop; a quadruple edge; too many loops; too many double edges; too many triple edges;
 * 		too many multiple edges (edges with multiplicity > 1, exact multiplicity undetermined upon rejection).
 * Otherwise, returns true.   																				*/
bool generatePairing(){

	int64_t	  ii, randNum;
	int 	  val1, val2;
	adjNode   *existing_node;
	int 	  locations_count = 0, multi_edges_count = 0;
	multiEdge *multi_edge_entry;

	// First initialize all relevant data structures.
	double_count = triple_count = 0;
	memset(hash_array, 0, sizeof(adjNode*) * n * hash_cells_per_vertex);
	for(ii = 0; ii < n; ii++){
		V[ii].single_edges_num = 0;
		END_OF_POOL(ii)	= START_OF_POOL(ii);
	}

	// loop_count and the points array still hold the information from the previous iteration.
	while(loop_count > 0){
		loop_count--;
		V[points[START_OF_SINGLE_EDGES]].has_loop = false;
	}

	// Generate the pairing.
	for(ii = M - 1; ii > 1; ii--){
		randNum = randInt(ii);
		ii--;
		POINTS_SWAP(ii, randNum);
	}

	printPairing();
	checkPointsArray();

	// Compress loops to the beginning of the points array; keep track of multiple edges; create adjacency nodes.
	for(ii = M - 2; ii >= START_OF_SINGLE_EDGES; ii -= 2){
		val1 = points[ii];
		val2 = points[ii + 1];

		if(val1 == val2){
			// A loop has been detected
			if(V[val1].has_loop){
				// Double loop at v_i. The pairing is invalid.
				LOG_DEBUG("double loop at %d.\n", val1);
				double_loop++;
				return false;
			}

			// This is a single loop. Check if the limit on loops is not exceeded.
			if(loop_count == B_L){
				// Too many loops. The pairing is invalid.
				LOG_DEBUG("too many loops (limit is %d).\n", B_L);
				many_loops++;
				return false;
			}

			// Compress all loops to the beginning of the points array.
			POINTS_SWAP(ii, START_OF_SINGLE_EDGES);
			POINTS_SWAP(ii + 1, START_OF_SINGLE_EDGES + 1);

			// Update loop count and data for the vertex.
			loop_count++;
			V[val1].has_loop = true;

			// Increase counter to handle the swapped pair in the next iteration.
			ii += 2;
		}
		else{
			// A non-loop is detected. Check whether v2 is already a neighbor of v1.
			GET_NODE(val1, val2, existing_node);
			if(existing_node == NULL){
				// First appearance of this edge. Create new single adjacency nodes for both vertices.
				addNewSingleNeighbor(val1, val2, ii);
				addNewSingleNeighbor(val2, val1, ii);

				// Update single edges count for both vertices.
				V[val1].single_edges_num++;
				V[val2].single_edges_num++;
			}
			else{
				// Multiple edge detected.
				if(existing_node->edge == 2 * M){
					// Quadruple edge. The pairing is invalid.
					LOG_DEBUG("quadruple edge at %d-%d.\n", val1, val2);
					quadruple_edge++;
					return false;
				}

				// Check whether this is the second or third appearance of the edge.
				if(existing_node->edge < M){
					// New multiple edge detected.
					if(multi_edges_count == B_D + B_T){
						// Too many multiple edges. The pairing is invalid.
						LOG_DEBUG("too many multiple edges (limit is %d).\n", B_D + B_T);
						many_multiple_edges++;
						return false;
					}

					// Create new multiple edge indicator.
					multi_edge_entry 		 = multi_edges + multi_edges_count;
					multi_edge_entry->v1	 = val1;
					multi_edge_entry->v2 	 = val2;
					multi_edge_entry->first  = existing_node->edge;
					multi_edge_entry->second = ii;
					multi_edge_entry->third  = -1;

					// Mark the locations of the two detected copies of the edge.
					multi_edges_locations[locations_count] 	   = existing_node->edge;
					multi_edges_locations[locations_count + 1] = ii;
					locations_count += 2;

					// Link the existing adjacency nodes of the edge to the new multiple edge entry.
					existing_node->edge = M + multi_edges_count;
					GET_NODE(val2, val1, existing_node);
					existing_node->edge = M + multi_edges_count;

					// Update single edges count for both vertices.
					V[val1].single_edges_num--;
					V[val2].single_edges_num--;

					// Update multiple edges count.
					multi_edges_count++;
				}
				else{
					// This is a triple edge.
					if(triple_count == B_T){
						// Too many triple edges. The pairing is invalid.
						LOG_DEBUG("too many triple edges (limit is %d).\n", B_T);
						many_triple_edges++;
						return false;
					}

					// Keep this third location in the multi-edge entry and in the list of multi-edges locations.
					multi_edges[existing_node->edge - M].third = ii;
					multi_edges_locations[locations_count] = ii;
					locations_count++;
					triple_count++;

					// Mark this edge as triple in the adjacency nodes.
					existing_node->edge = 2 * M;
					GET_NODE(val2, val1, existing_node);
					existing_node->edge = 2 * M;
				}
			}
		}
	}

	// Check if there are too many double edges.
	double_count = multi_edges_count - triple_count;
	if(double_count > B_D){
		// Too many double edges. The pairing is invalid.
		LOG_DEBUG("too many double edges (limit is %d).\n", B_D);
		many_double_edges++;
		return false;
	}

	// Compute P2.
	P2 = 0;
	for(ii = 0; ii < n; ii++){
		if(!V[ii].has_loop)
			P2 += FALL_2(V[ii].single_edges_num);
	}

	// Compress multiple edges to the end of the points array. First sort all locations of multiple edges.
	qsort(multi_edges_locations, locations_count, sizeof(type_M), type_M_Compare);

	/* Copy the pairs currently at the end of the points array to the locations of multiple edges.
	 * Since this is done "backwards" there is no problem if some multiple edges locations are in
	 * the cleared section. For single edges, the new location is updated in the adjacency nodes. */
	ii = M;
	while(locations_count > 0){
		ii -= 2;
		locations_count--;
		COPY_EDGE(ii, multi_edges_locations[locations_count]);
	}

	// Now keep all multiple edges in the designated sections.
	type_M double_pos = START_OF_DOUBLE_EDGES;
	type_M triple_pos = START_OF_TRIPLE_EDGES;

	while(multi_edges_count > 0){
		multi_edges_count--;
		multi_edge_entry = multi_edges + multi_edges_count;
		LOG_DEBUG("v1 %d, v2 %d, first %"PRINT_M", second %"PRINT_M", third %"PRINT_M"\n",
					multi_edge_entry->v1, multi_edge_entry->v2,
					multi_edge_entry->first/2, multi_edge_entry->second/2,
					multi_edge_entry->third == -1 ? -1 : multi_edge_entry->third/2);

		if(multi_edge_entry->third == -1){
			// This is a double edge.
			points[double_pos]		= points[double_pos + 2] = multi_edge_entry->v1;
			points[double_pos + 1] 	= points[double_pos + 3] = multi_edge_entry->v2;
			double_pos += 4;
		}
		else{
			// This is a triple edge.
			points[triple_pos]		= points[triple_pos + 2] = points[triple_pos + 4] = multi_edge_entry->v1;
			points[triple_pos + 1] 	= points[triple_pos + 3] = points[triple_pos + 5] = multi_edge_entry->v2;
			triple_pos += 6;
		}
	}

	printPairing();
	printAllNeighbors();
	verifyDataStructures();

	return true;
}

/* Attempts to perform an L-switch.
 * A random loop v2 and two random single edges (v1,v4) and (v3,v5) are
 * replaced (if possible) by the single edges (v1,v2), (v2,v3) and (v4,v5).
 * In case of vertex collision or forbidden edges returns false (f-rejection).
 * Otherwise, returns true and updates S to be the 2-path v1-v2-v3. 		*/
bool performL(int* S){

	int v1, v2, v3, v4, v5;
	u64 loop, single_edge1, single_edge2;

	// Selects a random loop and two random single edges.
	loop  		 = RANDOM_LOOP;
	single_edge1 = RANDOM_SINGLE_EDGE;
	single_edge2 = RANDOM_SINGLE_EDGE;


	/* Label all participating vertices and point edges to their left cells
	 * in the points array (the loop already points to its left cell). 	 */
	v2 = points[loop];
	ADJUST_EDGE(v1, v4, single_edge1);
	ADJUST_EDGE(v3, v5, single_edge2);

	LOG_DEBUG("Selection for L-Switch. Loop: %d (pair %"PRIu64"), "
			  "edge 1: %d-%d (pair %"PRIu64"), edge 2: %d-%d (pair %"PRIu64").\n",
			  v2, loop/2, v1, v4, single_edge1/2, v3, v5, single_edge2/2);

	// Check for vertex collision.
	if(v2 == v1 || v2 == v3 || v2 == v4 || v2 == v5 || v1 == v3 || v1 == v5 || v4 == v3 || v4 == v5){
		LOG_DEBUG("Vertex collision while attempting to perform L-Switch.\n");
		return false;
	}

	// Check for forbidden edges.
	if(IS_EDGE(v1,v2) || IS_EDGE(v2,v3) || IS_EDGE(v4,v5)){
		LOG_DEBUG("Forbidden edge(s) while attempting to perform L-Switch:%s%s%s.\n",
					IS_EDGE(v1,v2)? " v1-v2" : "", IS_EDGE(v2,v3)? " v2-v3" : "", IS_EDGE(v4,v5)? " v4-v5" : "");

		printNeighborsByHash(v1);
		printNeighborsByHash(v2);
		printNeighborsByHash(v3);
		printNeighborsByHash(v4);
		printNeighborsByHash(v5);
		return false;
	}

	// L-Switch can be performed. Update S.
	S[0] = v1;
	S[1] = v2;
	S[2] = v3;

	// Compress loops to the beginning of the points array.
	loop_count--;
	points[loop]	 = points[START_OF_SINGLE_EDGES];
	points[loop + 1] = points[START_OF_SINGLE_EDGES + 1];

	// Update new edges in the points array.
	points[single_edge1] 			  = v1;
	points[single_edge1 + 1] 		  = v2;
	points[single_edge2] 			  = v2;
	points[single_edge2 + 1] 		  = v3;
	points[START_OF_SINGLE_EDGES] 	  = v4;
	points[START_OF_SINGLE_EDGES + 1] = v5;

	// Mark v2 as not having a loop anymore and add its contribution to P2.
	V[v2].has_loop = false;
	V[v2].single_edges_num += 2;
	P2 += FALL_2(V[v2].single_edges_num);

	// Add v1 and v3 as new single neighbors of v2.
	addNewSingleNeighbor(v2, v1, single_edge1);
	addNewSingleNeighbor(v2, v3, single_edge2);

	// For the other vertices change adjacency nodes.
	changeNeighbor(v1, v4, v2, single_edge1);
	changeNeighbor(v3, v5, v2, single_edge2);
	changeNeighbor(v4, v1, v5, START_OF_SINGLE_EDGES);
	changeNeighbor(v5, v3, v4, START_OF_SINGLE_EDGES);

	LOG_DEBUG("S: [%d,%d,%d], e: (%d,%d).\n", v1, v2, v3, v4, v5);
	printPairing();
	printNeighborsByHash(v1);
	printNeighborsByHash(v2);
	printNeighborsByHash(v3);
	printNeighborsByHash(v4);
	printNeighborsByHash(v5);

	return true;
}

/* Attempts to perform a T-switch.
 * A random triple edge (v1,v2) and three random single edges (v3,v4), (v5,v6) and (v7,v8) are
 * replaced (if possible) by single edges (v1,v3), (v1,v5), (v1,v7), (v2,v4), (v2,v6) and (v2,v8).
 * In case of vertex collision or forbidden edges returns false (f-rejection).
 * Otherwise, returns true and updates S to be the 3-star [v1; v3,v5,v7]. 						*/
bool performT(int* S){

	int v1, v2, v3, v4, v5, v6, v7, v8;
	u64 triple_edge, single_edge1, single_edge2, single_edge3;

	// Select a random triple edge and three random single edges.
	triple_edge  = RANDOM_TRIPLE_EDGE;
	single_edge1 = RANDOM_SINGLE_EDGE;
	single_edge2 = RANDOM_SINGLE_EDGE;
	single_edge3 = RANDOM_SINGLE_EDGE;

	// Point the triple edge to its leftmost copy in the points array while maintaining orientation.
	triple_edge -= (((triple_edge - START_OF_TRIPLE_EDGES) % 6) / 2) * 2;

	// Label all participating vertices and point edges to their left cells in the points array.
	ADJUST_EDGE(v1, v2, triple_edge);
	ADJUST_EDGE(v3, v4, single_edge1);
	ADJUST_EDGE(v5, v6, single_edge2);
	ADJUST_EDGE(v7, v8, single_edge3);

	LOG_DEBUG("Selection for T-Switch. Triple edge: %d-%d (pair %"PRIu64"), "
			  "single edge 1: %d-%d (pair %"PRIu64"), single edge 2: %d-%d (pair %"PRIu64"), "
			  "single edge 3: %d-%d (pair %"PRIu64").\n", v1, v2, triple_edge/2,
			  v3, v4, single_edge1/2, v5, v6, single_edge2/2, v7, v8, single_edge3/2);

	// Check for vertex collision.
	if(v1 == v3 || v1 == v4 || v1 == v5 || v1 == v6 || v1 == v7 || v1 == v8 ||
	   v2 == v3 || v2 == v4 || v2 == v5 || v2 == v6 || v2 == v7 || v2 == v8 ||
	   v3 == v5 || v3 == v6 || v3 == v7 || v3 == v8 ||
	   v4 == v5 || v4 == v6 || v4 == v7 || v4 == v8 ||
	   v5 == v7 || v5 == v8 || v6 == v7 || v6 == v8){
		LOG_DEBUG("Vertex collision while attempting to perform T-Switch.\n");
		return false;
	}

	// Check for forbidden edges.
	if(IS_EDGE(v1,v3) || IS_EDGE(v1,v5) || IS_EDGE(v1,v7) ||
	   IS_EDGE(v2,v4) || IS_EDGE(v2,v6) || IS_EDGE(v2,v8)){
		LOG_DEBUG("Forbidden edge(s) while attempting to perform T-Switch:%s%s%s%s%s%s.\n",
					IS_EDGE(v1,v3)? " v1-v3" : "", IS_EDGE(v1,v5)? " v1-v5" : "",
					IS_EDGE(v1,v7)? " v1-v7" : "", IS_EDGE(v2,v4)? " v2-v4" : "",
					IS_EDGE(v2,v6)? " v2-v6" : "", IS_EDGE(v2,v8)? " v2-v8" : "");
		printNeighborsByHash(v1);
		printNeighborsByHash(v2);
		printNeighborsByHash(v3);
		printNeighborsByHash(v4);
		printNeighborsByHash(v5);
		printNeighborsByHash(v6);
		printNeighborsByHash(v7);
		printNeighborsByHash(v8);
		return false;
	}

	// T-Switch can be performed. Update S.
	S[0] = v1;
	S[1] = v3;
	S[2] = v5;
	S[3] = v7;

	// If v1/v2 was incident to r single edges its contribution to P2 is increased by (r+3)(r+2) - r(r-1) = 6r + 6.
	P2 += (6 * V[v1].single_edges_num + 6) + (6 * V[v2].single_edges_num + 6);

	// If v1/v2 was incident to r single edges its contribution to S3 is increased by (r+3)(r+2)(r+1) - r(r-1)(r-2) = 9r(r+1) + 6.
	S3 += (9 * FALL_2(V[v1].single_edges_num + 1) + 6) + (9 * FALL_2(V[v2].single_edges_num + 1) + 6);

	V[v1].single_edges_num += 3;
	V[v2].single_edges_num += 3;

	// Compress triple edges between single edges and double edges in the points array.
	for(int ii = 0; ii < 6; ii++){
		points[triple_edge + ii] = points[START_OF_TRIPLE_EDGES + ii];
	}

	// Update new edges in the points array.
	points[START_OF_TRIPLE_EDGES]		= v1;
	points[START_OF_TRIPLE_EDGES + 1]	= v3;
	points[START_OF_TRIPLE_EDGES + 2] 	= v1;
	points[START_OF_TRIPLE_EDGES + 3] 	= v5;
	points[START_OF_TRIPLE_EDGES + 4] 	= v1;
	points[START_OF_TRIPLE_EDGES + 5] 	= v7;
	points[single_edge1] 				= v2;
	points[single_edge1 + 1] 			= v4;
	points[single_edge2] 				= v2;
	points[single_edge2 + 1] 			= v6;
	points[single_edge3] 				= v2;
	points[single_edge3 + 1] 			= v8;

	// Update / create relevant adjacency nodes.
	// v1 and v2 have two unused nodes in their nodes pool due to the triple edge.
	addNewSingleNeighbor(v1, v3, START_OF_TRIPLE_EDGES);
	addNewSingleNeighbor(v1, v5, START_OF_TRIPLE_EDGES + 2);
	addNewSingleNeighbor(v2, v4, single_edge1);
	addNewSingleNeighbor(v2, v6, single_edge2);

	// For all other new single edges use existing nodes.
	changeNeighbor(v1, v2, v7, START_OF_TRIPLE_EDGES + 4);
	changeNeighbor(v2, v1, v8, single_edge3);
	changeNeighbor(v3, v4, v1, START_OF_TRIPLE_EDGES);
	changeNeighbor(v5, v6, v1, START_OF_TRIPLE_EDGES + 2);
	changeNeighbor(v7, v8, v1, START_OF_TRIPLE_EDGES + 4);
	changeNeighbor(v4, v3, v2, single_edge1);
	changeNeighbor(v6, v5, v2, single_edge2);
	changeNeighbor(v8, v7, v2, single_edge3);

	// Update triple count.
	triple_count--;

	LOG_DEBUG("S: [%d; %d,%d,%d], S': [%d; %d,%d,%d].\n", v1, v3, v5, v7, v2, v4, v6, v8);
	printPairing();
	printNeighborsByHash(v1);
	printNeighborsByHash(v2);
	printNeighborsByHash(v3);
	printNeighborsByHash(v4);
	printNeighborsByHash(v5);
	printNeighborsByHash(v6);
	printNeighborsByHash(v7);
	printNeighborsByHash(v8);

	return true;
}

/* Attempts to perform a Type I D-switch.
 * A random double edge (v2,v5) and two random single edges (v1,v4) and (v3,v6)
 * are replaced (if possible) by the edges (v1,v2), (v2,v3), (v4,v5) and (v5,v6).
 * If none of the new edges already exists (thus they all remain single) this is a Class A switch.
 * If exactly one of the new edges already exists, and is single (and thus becomes a double edge),
 * the switch is Class B and the vertices are relabeled so that the new double edge is (v2,v3).
 * In case of vertex collision or forbidden edges (i.e. not Class A or B) returns false (f-rejection).
 * Otherwise, returns true, updates the switch class, and updates S to be the 2-path v1-v2-v3.		*/
bool performDTypeI(int* S, bool* is_class_A){
	int v1, v2, v3, v4, v5, v6;
	u64 double_edge, single_edge1, single_edge2;

	// Select a random double edge and two random single edges.
	double_edge  = RANDOM_DOUBLE_EDGE;
	single_edge1 = RANDOM_SINGLE_EDGE;
	single_edge2 = RANDOM_SINGLE_EDGE;

	// Point the double edge to its leftmost copy in the points array while maintaining orientation.
	double_edge -= (((double_edge - START_OF_DOUBLE_EDGES) % 4) / 2) * 2;

	// Label all participating vertices and point edges to their left cells in the points array.
	ADJUST_EDGE(v2, v5, double_edge);
	ADJUST_EDGE(v1, v4, single_edge1);
	ADJUST_EDGE(v3, v6, single_edge2);

	LOG_DEBUG("Selection for Type I D-Switch. Double edge: %d-%d (pair %"PRIu64"), "
			  "single edge 1: %d-%d (pair %"PRIu64"), single edge 2: %d-%d (pair %"PRIu64").\n",
			  v2, v5, double_edge/2, v1, v4, single_edge1/2, v3, v6, single_edge2/2);

	// Check for vertex collision.
	if(v2 == v1 || v2 == v3 || v2 == v4 || v2 == v6 ||
	   v5 == v1 || v5 == v3 || v5 == v4 || v5 == v6 ||
	   v1 == v3 || v1 == v6 || v4 == v3 || v4 == v6){
		LOG_DEBUG("Vertex collision while attempting to perform a Type I D-Switch.\n");
		return false;
	}

	// Check for existing edges. If a double edge is found returns false.
	existing_edge_type existing_edge = no_edge;
	CHECK_FOR_EXISTING_EDGE(v2,v3);
	CHECK_FOR_EXISTING_EDGE(v2,v1);
	CHECK_FOR_EXISTING_EDGE(v5,v6);
	CHECK_FOR_EXISTING_EDGE(v5,v4);

	// Check if too many edges were detected.
	if(existing_edge == multiple_edges || (revert_to_inc_gen && existing_edge != no_edge)){
		LOG_DEBUG("Too many edges while attempting to perform type I D-Switch:%s%s%s%s.\n",
					IS_EDGE(v1,v2)? " v1-v2" : "", IS_EDGE(v2,v3)? " v2-v3" : "",
					IS_EDGE(v4,v5)? " v4-v5" : "", IS_EDGE(v5,v6)? " v5-v6" : "");
		printNeighborsByHash(v1);
		printNeighborsByHash(v2);
		printNeighborsByHash(v3);
		printNeighborsByHash(v4);
		printNeighborsByHash(v5);
		printNeighborsByHash(v6);
		return false;
	}

	// D-Switch can be performed. Check which class this switch is.
	if(existing_edge == no_edge)
		// No edges - Class A.
		*is_class_A = true;
	else{
		// This is a Class B switch.
		*is_class_A = false;

		// If necessary, relabel vertices such that the existing edge is (v2,v3).
		if(SWAP_SINGLE_EDGES){
			// Swap single edges.
			SIMPLE_SWAP(v1, v3);
			SIMPLE_SWAP(v4, v6);
		}
		if(REVERSE_ORIENTATION){
			// Reverse orientation of each edge.
			SIMPLE_SWAP(v1, v4);
			SIMPLE_SWAP(v2, v5);
			SIMPLE_SWAP(v3, v6);
		}
	}

	// Update S.
	S[0] = v1;
	S[1] = v2;
	S[2] = v3;

	// Update points array and change adjacency nodes for all vertices.
	// The edges (v4,v5) and (v5,v6) are treated the same way for both classes.
	points[single_edge1] 	 = v4;
	points[single_edge1 + 1] = v5;
	points[single_edge2] 	 = v5;
	points[single_edge2 + 1] = v6;
	changeNeighbor(v5, v2, v4, single_edge1);
	changeNeighbor(v4, v1, v5, single_edge1);
	changeNeighbor(v5, v2, v6, single_edge2);
	changeNeighbor(v6, v3, v5, single_edge2);

	// The other edges treatment depends on the class.
	if(*is_class_A == true){
		// Compress double edges to the end of the points array.
		for(int ii = 0; ii < 4; ii++){
			points[double_edge + ii] = points[START_OF_DOUBLE_EDGES + ii];
		}

		// Update new edges in the points array.
		points[START_OF_DOUBLE_EDGES]	  = v1;
		points[START_OF_DOUBLE_EDGES + 1] = v2;
		points[START_OF_DOUBLE_EDGES + 2] = v2;
		points[START_OF_DOUBLE_EDGES + 3] = v3;

		// Update adjacency nodes.
		changeNeighbor(v2, v5, v1, START_OF_DOUBLE_EDGES);
		changeNeighbor(v1, v4, v2, START_OF_DOUBLE_EDGES);
		changeNeighbor(v2, v5, v3, START_OF_DOUBLE_EDGES + 2);
		changeNeighbor(v3, v6, v2, START_OF_DOUBLE_EDGES + 2);

		// Update double count.
		double_count--;
	}
	else{
		// (v2,v3) is a new double edge.
		points[double_edge]		= points[double_edge + 2] = v2;
		points[double_edge + 1] = points[double_edge + 3] = v3;

		// The edge (v1,v2) has to be kept in the location of the old (v2,v3) copy.
		// Find this location in the points array.
		adjNode* node;
		GET_NODE(v2, v3, node);
		type_M v2_v3_old_edge = node->edge;

		// Update points array.
		points[v2_v3_old_edge]	   = v1;
		points[v2_v3_old_edge + 1] = v2;

		// Update adjacency nodes.
		changeNeighbor(v2, v5, v1, v2_v3_old_edge);
		changeNeighbor(v1, v4, v2, v2_v3_old_edge);
		changeNeighborToDouble(v2, v5, v3);
		changeNeighborToDouble(v3, v6, v2);
	}

	// If v5 was incident to r single edges its contribution to P2 is increased by (r+2)(r+1) - r(r-1) = 4r + 2.
	P2 += (4 * V[v5].single_edges_num + 2);
	V[v5].single_edges_num += 2;

	// The remaining change of P2 depends on the class.
	if(*is_class_A == true){
		// If v2 was incident to r single edges its contribution to P2 is increased by (r+2)(r+1) - r(r-1) = 4r + 2.
		P2 += (4 * V[v2].single_edges_num + 2);
		V[v2].single_edges_num += 2;
	}
	else{
		// If v3 was incident to r single edges its contribution to P2 is decreased by r(r-1) - (r-2)(r-3) = 4r - 6.
		P2 -= (4 * V[v3].single_edges_num - 6);
		V[v3].single_edges_num -= 2;
	}

	LOG_DEBUG("S: [%d,%d,%d], P: %d-%d-%d.\n", v1, v2, v3, v4, v5, v6);
	printPairing();
	printNeighborsByHash(v1);
	printNeighborsByHash(v2);
	printNeighborsByHash(v3);
	printNeighborsByHash(v4);
	printNeighborsByHash(v5);
	printNeighborsByHash(v6);
	return true;
}

/* Attempts to perform a Type II D-switch.
 * A random double edge (v2,v5) and two random single edges (v1,v4) and (v3,v6) are
 * selected, such that (v2,v1) and (v2,v3) are single edges. These edges are replaced
 * (if possible) by double edges (v1,v2) and (v2,v3), and single edges (v4,v5) and (v5,v6).
 * In case of vertex collision or forbidden edges returns false (f-rejection).
 * Otherwise, returns true and updates S to be the 2-path v1-v2-v3. 					 */
bool performDTypeII(int* S){
	int		v1, v2, v3, v4, v5, v6;
	u64		double_edge;
	adjNode *node_v1, *node_v3, *node_v4, *node_v6;

	// Select a random double edge.
	double_edge = RANDOM_DOUBLE_EDGE;

	// Label edge's vertices and point edge to the left cell of the left copy in the points array.
	double_edge -= (((double_edge - START_OF_DOUBLE_EDGES) % 4) / 2) * 2;
	ADJUST_EDGE(v2, v5, double_edge);

	// Select v1 and v3, two neighbors of v2 different from v5.
	selectRandomNeighborsOfDoubleEdge(v2, v5, &node_v1, &node_v3);
	v1 = node_v1->id;
	v3 = node_v3->id;

	// For both v1 and v3, select a random neighbor of the vertex different from v2.
	node_v4 = selectRandomNeighborsOfSingleEdge(v1, v2);
	node_v6 = selectRandomNeighborsOfSingleEdge(v3, v2);
	v4 = node_v4->id;
	v6 = node_v6->id;

	// Check if all chosen edges (i.e. (v2,v1), (v2,v3), (v1,v4) and (v3,v6)) are single.
	if(IS_NODE_OF_MULTI_EDGE(node_v1) || IS_NODE_OF_MULTI_EDGE(node_v3) ||
	   IS_NODE_OF_MULTI_EDGE(node_v4) || IS_NODE_OF_MULTI_EDGE(node_v6)){
		LOG_DEBUG("Selected neighbors for Type II D-switch are not all single.\n");
		return false;
	}

	LOG_DEBUG("Selection for Type II D-Switch. Double edge: %d-%d (pair %"PRIu64"), "
			  "single edge 1: %d-%d, single edge 2: %d-%d.\n", v2, v5, double_edge/2, v1, v4, v3, v6);

	// Check for vertex collision.
	if(v5 == v4 || v5 == v6 || v1 == v6 || v4 == v3 || v4 == v6){
		LOG_DEBUG("Vertex collision while attempting to perform a Type II D-Switch.\n");
		return false;
	}

	// Check for forbidden edges.
	if(IS_EDGE(v4,v5) || IS_EDGE(v5,v6)){
		LOG_DEBUG("Forbidden edge(s) while attempting to perform a Type II D-Switch:%s%s.\n",
					IS_EDGE(v4,v5)? " v4-v5" : "", IS_EDGE(v5,v6)? " v5-v6" : "");
		printNeighborsByHash(v4);
		printNeighborsByHash(v5);
		printNeighborsByHash(v6);
		return false;
	}

	// Type II D-Switch can be performed. Update S.
	S[0] = v1;
	S[1] = v2;
	S[2] = v3;

	// If v5 was incident to r single edges its contribution to P2 is increased by (r+2)(r+1) - r(r-1) = 4r + 2.
	P2 += (4 * V[v5].single_edges_num + 2);
	V[v5].single_edges_num += 2;

	// If v1/v2/v3 was incident to r single edges its contribution to P2 is decreased by r(r-1) - (r-2)(r-3) = 4r - 6.
	P2 -= (4 * V[v1].single_edges_num - 6) + (4 * V[v2].single_edges_num - 6) + (4 * V[v3].single_edges_num - 6);
	V[v1].single_edges_num -= 2;
	V[v2].single_edges_num -= 2;
	V[v3].single_edges_num -= 2;

	// Update double count.
	double_count++;

	// Make room in the points array for a new double edge.
	type_M locations[] = {node_v1->edge, node_v3->edge, node_v4->edge, node_v6->edge};
	qsort(locations, 4, sizeof(type_M), type_M_Compare);
	COPY_EDGE(START_OF_DOUBLE_EDGES + 2, locations[3]);
	COPY_EDGE(START_OF_DOUBLE_EDGES,     locations[2]);

	// Update new edges in the points array.
	points[START_OF_DOUBLE_EDGES]		= v2;
	points[START_OF_DOUBLE_EDGES + 1]	= v1;
	points[START_OF_DOUBLE_EDGES + 2] 	= v2;
	points[START_OF_DOUBLE_EDGES + 3] 	= v1;
	points[double_edge]					= v2;
	points[double_edge + 1]				= v3;
	points[double_edge + 2] 			= v2;
	points[double_edge + 3] 			= v3;
	points[locations[0]] 				= v4;
	points[locations[0] + 1] 			= v5;
	points[locations[1]] 				= v5;
	points[locations[1] + 1] 			= v6;

	// Update adjacency nodes for all vertices.
	changeNeighborToDouble(v2, v5, v1);
	changeNeighborToDouble(v1, v4, v2);
	changeNeighborToDouble(v2, v5, v3);
	changeNeighborToDouble(v3, v6, v2);
	changeNeighbor(v5, v2, v4, locations[0]);
	changeNeighbor(v4, v1, v5, locations[0]);
	changeNeighbor(v5, v2, v6, locations[1]);
	changeNeighbor(v6, v3, v5, locations[1]);

	LOG_DEBUG("S: [%d,%d,%d], P: %d-%d-%d.\n", v1, v2, v3, v4, v5, v6);
	printPairing();
	printNeighborsByHash(v1);
	printNeighborsByHash(v2);
	printNeighborsByHash(v3);
	printNeighborsByHash(v4);
	printNeighborsByHash(v5);
	printNeighborsByHash(v6);

	return true;
}

/* For a given simple 2-path S = v1-v2-v3, returns the number of ordered single edges e =(v4,v5),
 * such that e is vertex disjoint from S, and (v1,v4) and (v2,v5) are non-edges.
 * Done by counting the "bad" pairs and subtracting them from M, the total number of ordered pairs.	*/
u64 computeBLG(int* S){
	int		v1 = S[0], v2 = S[1], v3 = S[2], v4, v5;
	u64		ii, count;
	adjNode *node, *internal_node;

	// First count unordered pairs which do not correspond to simple edges.
	count = loop_count + 2 * double_count + 3 * triple_count;

	// Second, count unordered simple edges intersecting S.
	for(ii = 0; ii < 3; ii++){
		count += V[S[ii]].single_edges_num;
	}
	// Subtract simple edges with both endpoints in S which were double counted.
	if(IS_SINGLE_EDGE(v1, v3))
		count -= 3;
	else
		count -= 2;

	// The count so far was of unoriented edges, proceed with oriented edges.
	count *= 2;

	// Now count ordered simple edges (v4,v5) disjoint from S, such that a forbidden edge exists.

	// Iterate all neighbors v4 of v1 not in S.
	for(node = START_OF_POOL(v1); node < END_OF_POOL(v1); node++){
		v4 = node->id;
		if(v4 != v2 && v4 != v3){
			// Add to the count all single edges (v4,v5) such that v5 is not in S.
			count += V[v4].single_edges_num;
			for(ii = 0; ii < 3; ii++){
				if(IS_SINGLE_EDGE(v4, S[ii]))
					count--;
			}
		}
	}

	// Now iterate all neighbors v5 of v3.
	for(node = START_OF_POOL(v3); node < END_OF_POOL(v3); node++){
		v5 = node->id;
		if(v5 != v1 && v5 != v2){
			/* Instead of just counting the single edges incident to v5, check for each such
			 * edge whether its other endpoint is adjacent to v1 to avoid double counting. */
			for(internal_node = START_OF_POOL(v5); internal_node < END_OF_POOL(v5); internal_node++){
				v4 = internal_node->id;

				// Ignore multiple edges.
				if(IS_NODE_OF_MULTI_EDGE(internal_node))
					continue;

				// This is a single edge. Add to count if the other endpoint is not in S and not adjacent to v1.
				if(NOT_IN_S(v4) && IS_NO_EDGE(v4, v1))
					count++;
			}
		}
	}

#if SAFETY_CHECKS
	// Compute count by brute force.
	u64 brute_count = M;

	// Only count simple edges.
	for(ii = START_OF_SINGLE_EDGES; ii < START_OF_TRIPLE_EDGES; ii += 2){
		// Ignore edges intersecting S.
		if(NOT_IN_S(points[ii]) && NOT_IN_S(points[ii + 1])){
			// Check both orientations of the edge.
			if(IS_NO_EDGE(v1,points[ii]) && IS_NO_EDGE(v3,points[ii + 1]))
				brute_count--;
			if(IS_NO_EDGE(v3,points[ii]) && IS_NO_EDGE(v1,points[ii + 1]))
				brute_count--;
		}
	}

	if(count != brute_count){
		LOG_ERR("ERROR: count %"PRIu64", brute count %"PRIu64".\n", count, brute_count);
		printAllNeighbors();
		finalize(safety_check_failure);
	}

	LOG_DEBUG("count of bad pairs was confirmed by brute count.\n");
#endif /* SAFETY_CHECKS */

	return M - count;
}

/* For a given simple 3-star S = [v1; v3,v5,v7], returns the number of simple 3-stars S' = [v2; v4,v6,v8]
 * which are vertex disjoint from S, and such that (v1,v2), (v3,v4), (v5,v6) and (v7,v8) are non-edges.
 * Done by counting the "bad" 3-stars and subtracting them from S3, the total number of simple 3-stars.	*/
u64 computeBTG(int* S){
	int v1 = S[0], v3 = S[1], v5 = S[2], v7 = S[3], v2, v4, v6, v8;
	int ii, jj, single_edges_to_s, single_edges_not_to_s;
	int encountered_count = 0, actual_neighbors_v3, actual_neighbors_v5, actual_joint_v3_v5;
	u64 count = 0;
	adjNode *node, *internal_node;

	// First count stars intersecting S.
	for(ii = 0; ii < 4; ii++){
		// Count stars such that v2 is in S (and maybe v4, v6 or v8 as well).
		count += FALL_3(V[S[ii]].single_edges_num);

		// Count stars such that v2 is not in S. Iterate all single neighbors v2 of S[ii].
		for(node = START_OF_POOL(S[ii]); node < END_OF_POOL(S[ii]); node++){
			v2 = node->id;

			// Consider only single edges to a vertex v2 not in S.
			if(IS_NODE_OF_SINGLE_EDGE(node) && NOT_IN_S_3(v2)){
				/* Count the ordered 3-stars S' centered at v2 with v := S[ii] as a leaf.
				 * Each selection of 3 leaves (i.e. neighbors of v2) corresponds to 6 ordered stars.
				 * To avoid double counting the count is separated according to the number of leaves of S' in S.
				 *
				 * One leaf in S: choose two ordered leaves u,w not in S, and location (v4/v6/v8) for v.
				 * 2 leaves in S: choose a leaf u in S and a leaf w not in S, and choose location for w.
				 * 				  The other 3 ordered stars corresponding to {v,u,w} are counted when S[ii] = u.
				 * 3 leaves in S: choose two ordered leaves u,w from S.
				 * 				  Here only the case v = v4 is considered. The other 4 stars are counted when S[ii] = u,w.
				 */

				// First check how many of the single neighbors of v2 are in S.
				single_edges_to_s = 0;
				for(jj = 0; jj < 4; jj++){
					if(IS_SINGLE_EDGE(v2,S[jj]))
						single_edges_to_s++;
				}
				single_edges_not_to_s = V[v2].single_edges_num - single_edges_to_s;

				// Stars with exactly one leaf in S.
				count += 3 * FALL_2(single_edges_not_to_s);

				// Stars with exactly two leaves in S.
				count += 3 * (single_edges_to_s - 1) * single_edges_not_to_s;

				// Stars with all three leaves in S.
				count += FALL_2(single_edges_to_s - 1);
			}
		}
	}

	// Now count stars S' which are vertex disjoint from S, such that a forbidden edge exists.

	// First count stars such that (v1,v2) is an edge. Iterate all neighbors v2 of v1 not in S.
	for(node = START_OF_POOL(v1); node < END_OF_POOL(v1); node++){
		v2 = node->id;
		if(NOT_IN_S_3(v2)){
			// Count the single edges (v2,u) such that u is not in S.
			single_edges_not_to_s = V[v2].single_edges_num;
			for(ii = 0; ii < 4; ii++){
				if(IS_SINGLE_EDGE(v2, S[ii]))
					single_edges_not_to_s--;
			}
			// Every ordered 3-tuple of single neighbors of v2 (not to S) forms a distinct bad 3-star.
			count += FALL_3(single_edges_not_to_s);
		}
	}

	// Now count stars such that (v3,v4) is an edge but (v1,v2) is not. Iterate all neighbors v4 of v3 not in S.
	for(node = START_OF_POOL(v3); node < END_OF_POOL(v3); node++){
		v4 = node->id;
		if(NOT_IN_S_3(v4)){
			// Every single neighbor of v4 not in S is a potential star center v2.
			for(internal_node = START_OF_POOL(v4); internal_node < END_OF_POOL(v4); internal_node++){
				// Consider only single edges.
				if(IS_NODE_OF_MULTI_EDGE(internal_node))
					continue;

				// This is a single edge. Ignore if the other endpoint is in S or adjacent to v1.
				v2 = internal_node->id;
				if(NOT_IN_S_3(v2) && IS_NO_EDGE(v1, v2)){
					// The other two leaves of S' can be any ordered pair of single neighbors of v2 not in S.
					single_edges_not_to_s = V[v2].single_edges_num - 1;
					for(ii = 0; ii < 4; ii++){
						if(IS_SINGLE_EDGE(v2,S[ii]))
							single_edges_not_to_s--;
					}
					count += FALL_2(single_edges_not_to_s);

					// For double counting avoidance, keep count of the joint neighbors of v3 and (each encountered) v2.
					joint_neighbors_tracker[v2].joint_with_v3++;

					// Keep track of the encountered vertices v2 for later initialization.
					if(joint_neighbors_tracker[v2].joint_with_v3 == 1){
						encountered_count++;
						encountered[encountered_count] = v2;
					}
				}
			}
		}
	}

	// Now count stars such that (v5,v6) is an edge but not (v1,v2) or (v3,v4). Iterate all neighbors v6 of v5 not in S.
	for(node = START_OF_POOL(v5); node < END_OF_POOL(v5); node++){
		v6 = node->id;
		if(NOT_IN_S_3(v6)){
			// Every single neighbor of v6 not in S is a potential star center v2.
			for(internal_node = START_OF_POOL(v6); internal_node < END_OF_POOL(v6); internal_node++){
				// Consider only single edges.
				if(IS_NODE_OF_MULTI_EDGE(internal_node))
					continue;

				// This is a single edge. Ignore if the other endpoint is in S or adjacent to v1.
				v2 = internal_node->id;
				if(NOT_IN_S_3(v2) && IS_NO_EDGE(v1, v2)){
					// The other two leaves of S' can be any ordered pair of single neighbors of v2 not in S.
					single_edges_not_to_s = V[v2].single_edges_num - 1;
					for(ii = 0; ii < 4; ii++){
						if(IS_SINGLE_EDGE(v2,S[ii]))
							single_edges_not_to_s--;
					}
					count += FALL_2(single_edges_not_to_s);

					// Keep count of the joint neighbors of v5 and (the current) v2.
					joint_neighbors_tracker[v2].joint_with_v5++;

					// If v2 has joint neighbors with v3 we have to deal with double counting.
					if(joint_neighbors_tracker[v2].joint_with_v3 > 0){
						/* Every ordered pair (u,w) of single neighbors of v2 other than v6, such that
						 * u is a neighbor of v3, form with v2 and v6 a star that was already counted.
						 * Vertices v6 which are adjacent to both v3 and v5 require special care.	*/
						actual_neighbors_v3 = joint_neighbors_tracker[v2].joint_with_v3;
						if(IS_EDGE(v3,v6)){
							actual_neighbors_v3--;
							joint_neighbors_tracker[v2].joint_with_v3_v5++;
						}
						count -= actual_neighbors_v3 * (single_edges_not_to_s - 1);
					}
					else if(joint_neighbors_tracker[v2].joint_with_v5 == 1){
						// This is a newly encountered star center.
						encountered_count++;
						encountered[encountered_count] = v2;
					}
				}
			}
		}
	}

	// Finally, count stars such that (v7,v8) is an edge but not (v1,v2), (v3,v4) or (v5,v6).
	// Iterate all neighbors v8 of v7 not in S.
	for(node = START_OF_POOL(v7); node < END_OF_POOL(v7); node++){
		v8 = node->id;
		if(NOT_IN_S_3(v8)){
			// Every single neighbor of v8 not in S is a potential star center v2.
			for(internal_node = START_OF_POOL(v8); internal_node < END_OF_POOL(v8); internal_node++){
				// Consider only single edges.
				if(IS_NODE_OF_MULTI_EDGE(internal_node))
					continue;

				// This is a single edge. Ignore if the other endpoint is in S or adjacent to v1.
				v2 = internal_node->id;
				if(NOT_IN_S_3(v2) && IS_NO_EDGE(v1, v2)){
					// The other two leaves of S' can be any ordered pair of single neighbors of v2 not in S.
					single_edges_not_to_s = V[v2].single_edges_num - 1;
					for(ii = 0; ii < 4; ii++){
						if(IS_SINGLE_EDGE(v2,S[ii]))
							single_edges_not_to_s--;
					}
					count += FALL_2(single_edges_not_to_s);

					// If current v2 does not have any joint neighbors with v3 or v5 there is nothing left to do.
					if(joint_neighbors_tracker[v2].joint_with_v3 == 0 &&
					   joint_neighbors_tracker[v2].joint_with_v5 == 0)
						continue;

					// Remove v8 from the count of joint neighbors if adjacent to v3 or v5.
					actual_neighbors_v3 = joint_neighbors_tracker[v2].joint_with_v3;
					actual_neighbors_v5 = joint_neighbors_tracker[v2].joint_with_v5;
					actual_joint_v3_v5	= joint_neighbors_tracker[v2].joint_with_v3_v5;

					if(IS_EDGE(v3,v8))
						actual_neighbors_v3--;
					if(IS_EDGE(v5,v8)){
						actual_neighbors_v5--;
						if(actual_neighbors_v3 < joint_neighbors_tracker[v2].joint_with_v3)
							actual_joint_v3_v5--;
					}

					// Every star with two of the forbidden edges (including (v7,v8)) was already counted.
					count -= actual_neighbors_v3 * (single_edges_not_to_s - 1);
					count -= actual_neighbors_v5 * (single_edges_not_to_s - 1);

					// Every star with all three forbidden edges was just subtracted twice,
					// but should have only been subtracted once.
					count += actual_neighbors_v3 * actual_neighbors_v5 - actual_joint_v3_v5;
				}
			}
		}
	}

	// Clear all non-zero entries in joint_neighbors_tracker.
	while(encountered_count > 0){
		memset(joint_neighbors_tracker + encountered[encountered_count], 0, sizeof(jointNeighbors));
		encountered_count--;
	}

#if SAFETY_CHECKS
	// Compute count by brute force.
	u64 brute_count = S3;
	int *curr_edges;
	int curr_edges_num;
	ALLOCATE_MEMORY(curr_edges, int, degree);

	// Iterate all vertices not in S. Each is a potential star center v2.
	for(v2 = 0; v2 < n; v2++){
		if(!NOT_IN_S_3(v2))
			continue;

		// Keep all single neighbors of v2 not in S in a temporary array for easier iteration.
		curr_edges_num = 0;
		for(node = START_OF_POOL(v2); node < END_OF_POOL(v2); node++){
			if(NOT_IN_S_3(node->id) && IS_NODE_OF_SINGLE_EDGE(node)){
				curr_edges[curr_edges_num] = node->id;
				curr_edges_num++;
			}
		}

		// For each choice of distinct leaves v4, v6, v8, subtract the star from the count if no forbidden edges exist.
		for(v4 = 0; v4 < curr_edges_num; v4++){
			for(v6 = 0; v6 < curr_edges_num; v6++){
				if(v6 == v4)
					continue;
				for(v8 = 0; v8 < curr_edges_num; v8++){
					if(v8 == v4 || v8 == v6)
						continue;

					if(IS_NO_EDGE(v1, v2) && IS_NO_EDGE(v3, curr_edges[v4]) &&
					   IS_NO_EDGE(v5, curr_edges[v6]) && IS_NO_EDGE(v7, curr_edges[v8]))
						brute_count--;
				}
			}
		}
	}

	free(curr_edges);

	if(count != brute_count){
		LOG_ERR("ERROR: count %"PRIu64", brute count %"PRIu64".\n", count, brute_count);
		printAllNeighbors();
		finalize(safety_check_failure);
	}

	LOG_DEBUG("count of bad stars was confirmed by brute count.\n");

#endif /* SAFETY_CHECKS */

	return S3 - count;
}

/* For a given 2-path S = v1-v2-v3, where each of the edges (v1,v2) and (v2,v3) is either single or
 * double (according to the type and class of the D-switch), returns the number of simple 2-paths
 * v4-v5-v6 which are vertex disjoint from S, and such that (v1,v4), (v2,v5) and (v3,v6) are non-edges.
 * Done by counting the "bad" 2-paths and subtracting them from P2, the total number of simple 2-paths.  */
u64 computeBDG(int* S){
	int 	v1 = S[0], v2 = S[1], v3 = S[2], v4, v5, v6;
	int 	jj, single_neighbors, encountered_count = 0;
	u64 	ii, count = 0;
	adjNode *node, *internal_node;

	// First count paths v4-v5-v6 intersecting S.
	for(ii = 0; ii < 3; ii++){
		// Count paths such that v5 is in S (and maybe v4 or v6 as well).
		count += FALL_2(V[S[ii]].single_edges_num);

		// Count paths such that v5 is not in S. Iterate all single neighbors v5 of S[ii].
		for(node = START_OF_POOL(S[ii]); node < END_OF_POOL(S[ii]); node++){
			v5 = node->id;

			// Consider only single edges to a vertex v5 not in S.
			if(IS_NODE_OF_SINGLE_EDGE(node) && NOT_IN_S(v5)){
				/* The path can continue on any single edge incident to v5 but the one backwards to S[ii].
				 * For each path both directions are counted (i.e. S[ii] can be either v4 or v6).		*/
				count += 2 * (V[v5].single_edges_num - 1);

				// Avoid double counting if v4,v6 are both in S.
				for(jj = 0; jj < ii; jj++){
					if(IS_SINGLE_EDGE(v5,S[jj]))
						count -= 2;
				}
			}
		}
	}

	// Now count paths v4-v5-v6 disjoint from S, such that a forbidden edge exists.

	// First count paths such that (v2,v5) is an edge. Iterate all neighbors v5 of v2 not in S.
	for(node = START_OF_POOL(v2); node < END_OF_POOL(v2); node++){
		v5 = node->id;
		if(NOT_IN_S(v5)){
			// Avoid processing the same vertex twice if (v2,v5) is a double edge.
			if(IS_NEXT_NODE_THE_SAME(node))
				continue;

			// Count the single edges (v5,u) such that u is not in S.
			single_neighbors = V[v5].single_edges_num;
			for(jj = 0; jj < 3; jj++){
				if(IS_SINGLE_EDGE(v5, S[jj]))
					single_neighbors--;
			}
			// Any ordered pair of single neighbors forms a distinct bad 2-path.
			count += FALL_2(single_neighbors);
		}
	}

	// Now count paths v4-v5-v6 such that (v3,v6) is an edge but (v2,v5) is not. Iterate all neighbors v6 of v3 not in S.
	for(node = START_OF_POOL(v3); node < END_OF_POOL(v3); node++){
		v6 = node->id;
		if(NOT_IN_S(v6)){
			// Avoid processing the same vertex twice if (v3,v6) is a double edge.
			if(IS_NEXT_NODE_THE_SAME(node))
				continue;

			/* Now iterate all single neighbors v5 of v6 (not in S and not adjacent to v2),
			 * and for each count the single edges (v5,v4) for "new" v4. 				 */
			for(internal_node = START_OF_POOL(v6); internal_node < END_OF_POOL(v6); internal_node++){
				v5 = internal_node->id;

				// Consider only single edges.
				if(IS_NODE_OF_MULTI_EDGE(internal_node))
					continue;

				// This is a single edge. Ignore if the other endpoint is in S or adjacent to v2.
				if(NOT_IN_S(v5) && IS_NO_EDGE(v2, v5)){
					// The path can continue on any single edge incident to v5 but the one to v6 and those to S.
					count += V[v5].single_edges_num - 1;
					for(ii = 0; ii < 3; ii++){
						if(IS_SINGLE_EDGE(v5,S[ii]))
							count--;
					}

					// For double counting avoidance, keep count of the joint neighbors of v3 and (each encountered) v5.
					joint_neighbors_tracker[v5].joint_with_v3++;

					// Keep track of the encountered vertices v5 for later initialization.
					if(joint_neighbors_tracker[v5].joint_with_v3 == 1){
						encountered_count++;
						encountered[encountered_count] = v5;
					}
				}
			}
		}
	}


	// Finish by counting paths v4-v5-v6 such that (v1,v4) is an edge but not (v3,v6) or (v2,v5).

	// Iterate all neighbors v4 of v1 not in S.
	for(node = START_OF_POOL(v1); node < END_OF_POOL(v1); node++){
		v4 = node->id;
		if(NOT_IN_S(v4)){
			// Avoid processing the same vertex twice if (v1,v4) is a double edge.
			if(IS_NEXT_NODE_THE_SAME(node))
				continue;

			/* Now iterate all single neighbors v5 of v4 (not in S and not adjacent to v2),
			 * and for each count the single edges (v5,v6) for "new" v6 not adjacent to v3. */
			for(internal_node = START_OF_POOL(v4); internal_node < END_OF_POOL(v4); internal_node++){
				v5 = internal_node->id;

				// Consider only single edges.
				if(IS_NODE_OF_MULTI_EDGE(internal_node))
					continue;

				// This is a single edge. Ignore if the other endpoint is in S or adjacent to v2.
				if(NOT_IN_S(v5) && IS_NO_EDGE(v2, v5)){
					// The path can continue on any single edge incident to v5 but the one to v4 and those to S.
					count += V[v5].single_edges_num - 1;
					for(ii = 0; ii < 3; ii++){
						if(IS_SINGLE_EDGE(v5,S[ii]))
							count--;
					}

					/* It remains to subtract paths v4-v5-v6 such that both (v1,v4) and (v3,v6) are edges.
					 * Every such path was already counted once for each joint neighbor of v5 and v3.	*/
					count -= joint_neighbors_tracker[v5].joint_with_v3;

					// Final adjustment: if (v3,v4) is an edge then the non-path v4-v5-v4 was just subtracted.
					if(IS_EDGE(v3,v4))
						count++;
				}
			}
		}
	}

	// Set all positive values in joint_neighbors_tracker back to 0.
	while(encountered_count > 0){
		joint_neighbors_tracker[encountered[encountered_count]].joint_with_v3 = 0;
		encountered_count--;
	}

#if SAFETY_CHECKS
	// Compute count by brute force.
	u64 brute_count = P2;

	// Only count simple edges (v4,v5).
	for(ii = 0; ii < START_OF_DOUBLE_EDGES; ii += 2){
		v4 = points[ii];
		v5 = points[ii + 1];

		// Ignore edges intersecting S.
		if(!NOT_IN_S(v4) || !NOT_IN_S(v5))
			continue;

		// Ignore edges (v4,v5) with one of the forbidden edges (v1,v4) or (v2,v5).
		if(IS_NO_EDGE(v1,v4) && IS_NO_EDGE(v2,v5)){
			// Iterate all neighbors v6 of v5.
			for(node = START_OF_POOL(v5); node < END_OF_POOL(v5); node++){
				v6 = node->id;
				// Subtract from the count if v6 is a single neighbor of v5 other than v4,
				// not in S, and not adjacent to v3.
				if(IS_NODE_OF_SINGLE_EDGE(node) && v6 != v4 && NOT_IN_S(v6) && IS_NO_EDGE(v3,v6))
					brute_count--;
			}
		}

		// Check the other orientation of the edge in the same manner.
		v4 = points[ii + 1];
		v5 = points[ii];
		if(IS_NO_EDGE(v1,v4) && IS_NO_EDGE(v2,v5)){
			for(node = START_OF_POOL(v5); node < END_OF_POOL(v5); node++){
				v6 = node->id;
				if(IS_NODE_OF_SINGLE_EDGE(node) && v6 != v4 && NOT_IN_S(v6) && IS_NO_EDGE(v3,v6))
					brute_count--;
			}
		}
	}


	if(count != brute_count){
		LOG_ERR("ERROR: count %"PRIu64", brute count %"PRIu64".\n", count, brute_count);
		printAllNeighbors();
		finalize(safety_check_failure);
	}

	LOG_DEBUG("count of bad 2-paths was confirmed by brute count.\n");
#endif /* SAFETY_CHECKS */

	return P2 - count;
}

// Attempts to remove all loops one after the other. Returns false upon f/b-rejection, and true otherwise.
bool noLoops(){
	int S[3];
	u64 blg;
	u64 blm0 = M2 - (8 * degree * double_count) - (degree * degree * loop_count);
	u64 blm1 = M - (6 * degree * degree) + (4 * degree);

	total_noloops++;
	total_loop_num += loop_count;

	while(loop_count > 0){
		// Attempt to perform an L-switch.
		total_L_switch_calls++;
		if(!performL(S)){
			LOG_DEBUG("Forward L-Switch failed.\n");
			L_f_rej++;
			return false;
		}

		// Check backwards rejection.
		blm0 += degree * degree;
		blg = computeBLG(S);
		LOG_DEBUG("blm0 %"PRIu64", blm1 %"PRIu64", P2 %"PRIu64", blg %"PRIu64".\n", blm0, blm1, P2, blg);
		// Check b-rejection incrementally (always possible since blg <= P2 <= M2 < RAND_INT_MAX).
		if(checkBackwardsRejectionIncremental(blm0, blm1, P2, blg)){
			LOG_DEBUG("Loop backwards rejection.\n");
			L_b_rej++;
			return false;
		}
		LOG_DEBUG("One loop iteration completed.\n");
	}

	LOG_DEBUG("All loops removed successfully!\n");

	return true;
}

// Attempts to remove all triple edges one after the other. Returns false upon f/b-rejection, and true otherwise.
bool noTriples(){
	int  ii, S[4];
	u64  btg;
	u64  btm0 = M3 - degree * degree * (12 * double_count + 18 * triple_count);
	u64  btm1 = M3 - degree * degree * (12 * double_count + 18 * triple_count + 16 * degree + degree * degree);
	bool rejection;

	total_notriples++;
	total_triple_edges_num += triple_count;

	// Compute S3.
	S3 = 0;
	for(ii = 0; ii < n; ii++)
		S3 += FALL_3(V[ii].single_edges_num);

	while(triple_count > 0){
		// Attempt to perform a T-switch.
		total_T_switch_calls++;
		if(!performT(S)){
			LOG_DEBUG("Forward T-Switch failed.\n");
			T_f_rej++;
			return false;
		}

		// Check backwards rejection.
		btm0 += 18 * degree * degree;
		btm1 += 18 * degree * degree;
		btg = computeBTG(S);
		LOG_DEBUG("btm0 %"PRIu64", btm1 %"PRIu64", S3 %"PRIu64", btg %"PRIu64".\n", btm0, btm1, S3, btg);

		if(S3 > RAND_INT_MAX){
			// S3 is too large, needs to approximate.
			rejection = (randReal() >= (1.0 * btm0 * btm1) / (1.0 * S3 * btg));
		}
		else{
			// Check b-rejection incrementally (possible since btg <= S3 <= RAND_INT_MAX).
			rejection = checkBackwardsRejectionIncremental(btm0, btm1, S3, btg);
		}

		if(rejection){
			LOG_DEBUG("Triple edge backwards rejection.\n");
			T_b_rej++;
			return false;
		}

		LOG_DEBUG("One triple edge iteration completed.\n");
	}

	LOG_DEBUG("All triple edges removed successfully!\n");

	return true;
}

/* Attempts to remove all double edges one after the other, INC-REG version.
 * Returns false upon f/b-rejection, and true otherwise.				  */
bool noDoublesReg(){
	int    S[3];
	u64    bdg;
	u64    bA  = M2 * (M2 - (degree - 1) * (16 * double_count + 3 * degree * degree + degree + 6));
	u64    bB1 = M2 - (4 * double_count * degree + 3 * degree * degree * (3 + degree));
	bool   is_class_A, rejection;
	double rand_val;

	while(double_count > 0){
		total_D_switch_calls++;

		// Select switch type.
		rand_val = randReal();

		// Perform a Type I switch with probability 1 - epsilon.
		if(rand_val >= epsilon){
			LOG_DEBUG("Type I switch.\n");
			total_Type_I++;
			// Attempt to perform a Type I switch.
			if(!performDTypeI(S, &is_class_A)){
				LOG_DEBUG("Forward Type I D-Switch failed.\n");
				type_I_f_rej++;
				return false;
			}
			if(is_class_A){
				total_Class_A++;
				bA  += 16 * M2 * (degree - 1);
				bB1 += 4 * degree;
			}
			else
				total_Type_I_Class_B++;
		}
		else{
			// Perform a Type II switch with probability rhoII(double_count).
			if(rand_val < rhoII[double_count]){
				total_Type_II++;
				// Attempt to perform a Type II switch.
				LOG_DEBUG("Type II switch.\n");
				if(!performDTypeII(S)){
					LOG_DEBUG("Forward Type II D-Switch failed.\n");
					type_II_f_rej++;
					return false;
				}

				is_class_A = false;
				bA  -= 16 * M2 * (degree - 1);
				bB1 -= 4 * degree;
			}
			else{
				// Type-reject with probability 1 - epsilon - rhoII(double_count).
				LOG_DEBUG("type-rejection.\n");
				type_rej++;
				return false;
			}
		}

		// Check backwards rejection according to the class.
		bdg = computeBDG(S);
		if(is_class_A){
			// Check for b-rejection either precisely or by approximation, according to the value of M2.
			if(bA_approx){
				// P2 * bdg is too large, needs to approximate.
				rejection = (randReal() >= bA_approx[double_count] / (1.0 * P2 * bdg));
			}
			else{
				// P2 * bdg <= RAND_INT_MAX, use precise computation.
				LOG_DEBUG("bA %"PRIu64", P2 %"PRIu64", bdg %"PRIu64".\n", bA, P2, bdg);
				rejection = checkBackwardsRejectionNonIncremental(bA, P2 * bdg);
			}
		}
		else{
			// Use precise computation for b-rejection (always possible since bdg <= M2 < RAND_INT_MAX).
			LOG_DEBUG("bB1 %"PRIu64", bdg %"PRIu64".\n", bB1, bdg);
			rejection = checkBackwardsRejectionNonIncremental(bB1, bdg);
		}

		if(rejection){
			LOG_DEBUG("Double edge backwards rejection.\n");
			if(is_class_A)
				Class_A_b_rej++;
			else
				Class_B_b_rej++;
			return false;
		}
		LOG_DEBUG("One double edge iteration completed.\n");
	}

	LOG_DEBUG("All double edges removed successfully!\n");

	return true;
}

/* Attempts to remove all double edges one after the other, INC-GEN version.
 * Returns false upon f/b-rejection, and true otherwise.				  */
bool noDoublesGen(){
	int  S[3];
	u64  bdg;
	u64  bdm0 = M2 - (8 * degree * double_count);
	u64  bdm1 = M2 - (4 * double_count * (2 * degree - 3)) - (3 * degree * degree * degree);
	bool is_class_A_dummy;

	while(double_count > 0){
		// Attempt to perform a D-switch.
		total_D_switch_calls++;
		if(!performDTypeI(S, &is_class_A_dummy)){
			LOG_DEBUG("Forward D-Switch failed.\n");
			D_f_rej++;
			return false;
		}

		// Check backwards rejection.
		bdm0 += 8 * degree;
		bdm1 += 8 * degree - 12;
		bdg = computeBDG(S);

		// Check b-rejection incrementally (always possible since bdg <= P2 <= M2 < RAND_INT_MAX).
		LOG_DEBUG("bdm0 %"PRIu64", bdm1 %"PRIu64", P2 %"PRIu64", bdg %"PRIu64".\n", bdm0, bdm1, P2, bdg);
		if(checkBackwardsRejectionIncremental(bdm0, bdm1, bdg, P2)){
			LOG_DEBUG("Double edge backwards rejection.\n");
			D_b_rej++;
			return false;
		}
		LOG_DEBUG("One double edge iteration completed.\n");
	}

	LOG_DEBUG("All double edges removed successfully!\n");

	return true;
}

// Creates second adjacency nodes for all double edges and proceeds to the correct version of noDoubles.
bool noDoubles(){
	total_nodoubles++;
	total_double_edges_num += double_count;

	/* Create a second adjacency node for both endpoints of each double edge.
	 * This is done for easier updates of the data structures after Class B switches.
	 * Although in the INC-GEN version of noDoubles no Class B switches can occur, this
	 * is done anyway for consistency (and some data structures updates rely on this).	*/
	for(type_M ii = START_OF_DOUBLE_EDGES; ii < M; ii += 4){
		CREATE_SECOND_NODE(points[ii], points[ii + 1]);
		CREATE_SECOND_NODE(points[ii + 1], points[ii]);
	}

	// Proceed to the correct version of noDoubles.
	if(revert_to_inc_gen)
		return noDoublesGen();
	else
		return noDoublesReg();
}

// Main program. Verifies input and configuration, generates random graphs, and outputs them.
int main(int argc, char** argv) {

	int  ii, number_of_copies;
	bool count_failed_attempts = CONSECUTIVE_FAILED_ATTEMPTS_LIMIT;

	int total_attempts = 0, generation_fails = 0, noLoops_fails = 0,
		noTriples_fails = 0, noDoubles_fails = 0, consecutive_failed_attempts = 0;

	// First check if the program was run with the help option.
	if(argc == 2 && (!strcmp(argv[1], "h") || !strcmp(argv[1], "-h") || !strcmp(argv[1], "--h"))){
		printInputFormat();
		exit(user_termination);
	}

	// Verify the number of arguments given to the program is in range.
	if(argc < 2 || argc > 4){
		LOG_ERR("\nInvalid arguments given to the program!\n");
		printInputFormat();
		exit(input_failure);
	}

	// Determine input type.
	bool is_input_in_file = checkInputType(argc, argv);

	// Open log file.
	char* log_file_name = (is_input_in_file && (argc == 4)) ? argv[3] : DEFAULT_LOG_FILE;
	log_file = fopen(log_file_name, "w");
	if(!log_file){
		LOG_ERR("\nCould not open log file. Exiting.\n");
		exit(file_failure);
	}

	// Open output file if provided.
	if(is_input_in_file && argc > 2){
		output_file = fopen(argv[2], "w");
		if(!output_file){
			LOG_ERR("\nCould not open output file. Exiting.\n");
			fclose(log_file);
			exit(file_failure);
		}
	}
	else{
		// stdout is the default output stream if no output file was provided.
		output_file = stdout;
	}

	// Verify there is not any faulty user configuration.
	checkConfiguration(argv);

	// Get input and verify it is valid.
	if(is_input_in_file){
		// Extract input from the input file.
		char num_copies_str[12], n_str[12], d_str[12], *input_strings[] = {num_copies_str, n_str, d_str};
		readInputFile(argv[1], input_strings);

		// Check input.
		verifyInput(input_strings, &number_of_copies);
	}
	else{
		// Check input as stored in argv.
		verifyInput(argv + 1, &number_of_copies);
	}

	// Allocate memory and create all data structures which only depend on the degree sequence itself.
	initialize();

	if(DO_NOT_OUTPUT)
		LOG_INFO("\n---------- DO_NOT_OUTPUT flag is turned on. No graphs output. ----------\n\n");

	for(ii = 1; ii <= number_of_copies; ii++){
		LOG_DEBUG("\n\n------------------------------ Starting graph %d ------------------------------ \n\n", ii);
		if(SHOW_PROGRESS)
			LOG_ERR("%d\n", ii);

		// Loop until a simple graph is obtained.
		while(true){
			// Generate pairings until a valid one is obtained.
			while(true){
				// Alert user if many consecutive failed attempts occur.
				if(count_failed_attempts && consecutive_failed_attempts >= CONSECUTIVE_FAILED_ATTEMPTS_LIMIT){
					LOG_ERR("The last %d attempts to generate a graph failed.\n", consecutive_failed_attempts);
					PRINT_OVERALL_STATS(ERR);
					approveContinuation("Continue anyway (such message will not be prompted again)?");
					count_failed_attempts = false;
				}
				total_attempts++;
				if(generatePairing())
					break;
				generation_fails++;
				consecutive_failed_attempts++;
			}

			// Attempt to remove all loops if there are any. Start over upon failure.
			if(B_L > 0){
				if(loop_count > 0){
					if(!noLoops()){
						LOG_DEBUG("noLoops failed!\n");
						noLoops_fails++;
						consecutive_failed_attempts++;
						continue;
					}

					LOG_DEBUG("noLoops succeeded!\n");
				}
				else
					LOG_DEBUG("There are no loops to remove.\n");
			}

			verifyDataStructures();

			// Attempt to remove all triple edges if there are any. Start over upon failure.
			if(B_T > 0){
				if(triple_count > 0){
					if(!noTriples()){
						LOG_DEBUG("noTriples failed!\n");
						noTriples_fails++;
						consecutive_failed_attempts++;
						continue;
					}

					LOG_DEBUG("noTriples succeeded!\n");
				}
				else
					LOG_DEBUG("There are no triple edges to remove.\n");
			}

			verifyDataStructures();

			// Attempt to remove all double edges if there are any. Start over upon failure.
			if(B_D > 0){
				if(double_count > 0){
					if(!noDoubles()){
						LOG_DEBUG("noDoubles failed!\n");
						noDoubles_fails++;
						consecutive_failed_attempts++;
						continue;
					}

					LOG_DEBUG("noDoubles succeeded!\n");
				}
				else
					LOG_DEBUG("There are no double edges to remove.\n");
			}

			verifyDataStructures();
			LOG_DEBUG("Graph %d produced!\n", ii);
			consecutive_failed_attempts = 0;
			break;
		}

		// A simple graph has been obtained. Print it unless graphs output is disabled.
		outputGraph();
	}  /* End of main loop. */

	LOG_DEBUG("Finished producing all %d graphs.\n", number_of_copies);

	// Print statistics unless disabled.
	if(PRINT_SUMMARY){
		LOG_INFO("\n----------------------------------------------------------- Summary "
				 "-----------------------------------------------------------\n\n");

		LOG_INFO("Number of graphs = %d, n = %d, degree = %d, "
				 "Loops limit = %d, Double edges limit = %d, Triple edges limit = %d.\n",
				 number_of_copies, n, (int) degree, B_L, B_D, B_T);

		PRINT_OVERALL_STATS(INFO);
		printStatisticsPairings(generation_fails, total_attempts);
		printStatisticsLoops();
		printStatisticsTriples();
		printStatisticsDoubles();
	}

	finalize(successful_termination);
}
